#
#  Copyright 2001 Nathaniel E. Bray/Deekoo L.
#
#  See LICENSE for details.
#
package Yeemp::Client;
use Yeemp::Common;
use Yeemp::Channel;
use POSIX ':sys_wait_h',':fcntl_h';
use Fcntl ':flock';
use IPC::Open3;
use IPC::Open2;
require Exporter;
if (eval('require "asm/ioctls.ph";')) {
  my $FIONREAD = &FIONREAD;
} elsif (eval('require "mach/sys/filio.ph";')) {
  my $FIONREAD = &FIONREAD;
} else {
  if (!defined($FIONREAD=&slurp_define('FIONREAD',
    '/usr/include/asm/ioctls.h'))) {
    &UI::stderr("Unable to find FIONREAD ioctl.\n");
  }
}
#use strict;
#use warnings;

sub untaint_environment()
{
  $ENV{IFS}=' ';
  delete($ENV{BASH_ENV});
}

sub shutdown_client()
{
  save_contact_list();
  save_client_cfg();
  channel_close(CHANNEL);
  &UI::shutdown_ui();
  exit;
}

sub request_media($$)
{
  # Arg 0: User to request file from.
  # Arg 1: Filename, without leading /
  local *FILE;
  if (!sysopen(FILE,"$clientdir/cache/$_[0].dat",O_RDWR|O_CREAT)) {
    UI::stderr("Unable to open $clientdir/$_[0].dat\n");
    return;
  }
  if (!flock(FILE,LOCK_EX|LOCK_NB)) {
    UI::stderr("Error locking $clientdir/$_[0].dat\n");
  }
  local $/ = "\n";
  my @cachecontrol = <FILE>;
  chomp(@cachecontrol);
  local %cachecontrol;
  if (!eval('%cachecontrol = @cachecontrol; 1;')) {
    UI::stderr("Error slurping cache.\n");
  }
  if (!(exists($cachecontrol{$_[1]}) && ($cachecontrol{$_[1]}>time()-864000))) {
    deliver_message($_[0],
      "From: $Yeemp::Config::my_address\012Get: /$_[1]\012\012",1600000);
  }
  $cachecontrol{$_[1]} = time();
  seek(FILE,0,SEEK_SET) or UI::stderr("Seek error.\n");
  print FILE join("\n",%cachecontrol),"\n";
  truncate(FILE,tell(FILE));
  close(FILE);
}

sub media_file($$)
{
  # Arg 0: user to request from
  # Arg 1: File to request, sans leading /
  &UI::stderr("Looking for $clientdir/cache/$_[0]/$_[1]\n");
  if (-f "$clientdir/cache/$_[0]/$_[1]") {
    &UI::stderr("Found it.\n");
    return "$clientdir/cache/$_[0]/$_[1]";
  }
  if ($Yeemp::Config::request_media_files) {
    request_media($_[0], $_[1]);
  }
  if (-f "$clientdir/$_[1]") {
    return "$clientdir/$_[1]";
  } elsif (-f "/usr/local/share/yeemp/$_[1]") {
    return "/usr/local/share/yeemp/$_[1]";
  } elsif (-f "/usr/share/yeemp/$_[1]") {
    return "/usr/share/yeemp/$_[1]";
  }
  return undef;
}

sub play_ogg($)
{
  if ($Yeemp::Config::use_esd && (-S '/tmp/.esd/socket')) {
    return(system('esddsp', 'ogg123', $_[0])==0);
  }
  return (system('ogg123', $_[0])==0);
}

sub play_statchange($$)
{
  if ($Yeemp::Config::play_oggs) {
    my $flooble = media_file($_[0], "statsounds/$_[1].ogg");
    if (defined($flooble)) {
      play_ogg($flooble) and return;
    }
  }
  if ($Yeemp::Config::beep_on_statchange) {
    UI::statchange_beep();
  }
}

sub check()
{
  if (!defined($check_interval)) {
    $check_interval=0;
    $check_last=0;
  }
  local $now=time();
  if (!channel_is_open(CHANNEL)) {
    if (&open_channel()) {
      if (!&login_to_server()) {
        return 0;
      }
    } else {
      return 0;
    }
  }
  if ($Yeemp::Channel::NEW_received{CHANNEL}>0) {
	  $Yeemp::Channel::NEW_received{CHANNEL}--;
    return 1;
  }
  if ($Yeemp::Config::explicit_check) {
    local @resp=channel_command(CHANNEL,"Check");
    if (!defined($resp[0])) {
      &UI::stderr("Check failed.\n");
      channel_close(CHANNEL);
    } elsif ($resp[0] eq '1') {
      $check_interval=0;
      return 1;
    } elsif ($check_interval<29) {
      $check_interval++;
    }
  }
  $check_last=$now;
  return 0;
}

sub encrypt_message_to_tempkey
{
  if (!exists($contacts_sendfrom{$_[0]})) {
    &UI::stderr("Warning: contacts_sendfrom{$_[0]} is blank.\n");
    return undef;
  }
  local *WFH;
  local *RFH;
  local *EFH;
#  local $pid=
  open3(\*WFH, \*RFH, \*EFH, 'gpg', '--no-default-keyring',
    '--keyring', "$clientdir/temp-keyring", '--always-trust', '-e',
    '--recipient', "$_[0] (from $contacts_sendfrom{$_[0]})",
    @extra_gpg_options, '-o', '-');
  print WFH $_[1];
  close(WFH);
  local($/);
  local $mess=<RFH>;
  close(RFH);
  local $err=<EFH>;
  if (defined($err) && ($err ne '') && ($Yeemp::Config::debug)) {
    &UI::stderr("Debug: encrypt_message_to_tempkey: Errors: \n$err\n");
  }
  close(EFH);
  waitpid(-1,WNOHANG);
  return $mess;
}

sub delete_postponed_messages()
{
  local $err=0;
  local $copy=1;
  local *FILE;
  local *FILE2;
  if (open(FILE,"<$_[0]")) {
    if (flock(FILE,LOCK_EX)) {
      if (sysopen(FILE2,"$_[0].new",O_RDWR|O_CREAT|O_EXCL)) {
        if (flock(FILE2,LOCK_EX)) {
          local $ok=1;
          while ($line=<FILE>) {
            # This will need to change when Expiry headers go into postponed
            # messages
            if ($line=~/^To: \Q$_[1]\E$/) {
              $copy=0;
            } elsif ($line=~/^[^ ]/) {
              $copy=1;
            }
            if ($copy) {
              $ok &= print FILE2 $line;
            }
          }
          if ($ok) {
            truncate(FILE2,tell(FILE2));
            if (!rename("$_[0].new",$_[0])) {
              &UI::stderr("Error replacing $_[0]\n");
              $err=1;
            }
          } else {
            &UI::stderr("Error writing $_[0].new\n");
            $err=1;
          }
        } else {
          &UI::stderr("Error locking $_[0].new\n");
          $err=1;
        }
      } else {
        &UI::stderr("Error opening $_[0].new\n");
        $err=1;
      }
    } else {
      &UI::stderr("Error locking $_[0]\n");
      $err=1;
    }
    close(FILE);
  } else {
     &UI::stderr("Error opening $_[0]\n");
     $err=1;
  }
  if ($err) {
    &UI::stderr("Some postponed messages for $_[1] may not have been deleted.\n");
  }
}

sub purge_postponed()
{
  &delete_postponed_messages($clientdir.'postponed_bare',$_[0]);
  &delete_postponed_messages($clientdir.'postponed_crypted',$_[0]);
}

sub purge_contact()
{
  # Destroy a contact.
  local $address=$_[0];
  delete($contacts_presence{$address});
  if (exists($contacts_recvfrom{$address})) {
    system('gpg', '--no-default-keyring',
      '--secret-keyring', "${clientdir}secring",
      '--keyring', "${clientdir}pubring", '--yes',
      '--delete-secret-and-public-key',
      "$Yeemp::Config::my_address (from $contacts_recvfrom{$address})");
  }
  if (exists($contacts_sendfrom{$address})) {
    system('gpg', '--no-default-keyring',
      '--keyring', "${clientdir}pubring", '--yes',
      '--delete-key', "$address (from $contacts_sendfrom{$address})",
			'--status-fd', 2);
  }
  delete($contacts_sendfrom{$address});
  delete($contacts_recvfrom{$address});
  delete($contacts{$address});
  delete($contacts_status{$address});
  delete($contacts_shortnames{$address});
  &purge_postponed($address);
}

sub clean()
{
  local $text=$_[0];
  if (!defined($text)) {
    return '';
  }
  while ($text=~/([\x00-\x09\x0B-\x1F])/s) {
    local $match=$1;
    local $rep=sprintf('<%02X>',ord($match));
    $text=~s/\Q$match\E/$rep/sg;
  }
  $text=~s/\012/\n/s;
  return $text;
}

sub clean_print($)
{
  local $text=$_[0];
  while ($text=~/([\x00-\x09\x0B-\x1F\x80-\xFF])/s) {
    local $match=$1;
    local $rep=sprintf('<%02X>',ord($match));
    $text=~s/$match/$rep/s;
  }
  $text=~s/\012/\n/s;
  &UI::stdout($text);
}

sub slurp_until_eof()
{
  local @files=@_;
  local @filenos;
  local $rfds='';
  while (@files) {
    local $file=shift(@files);
    vec ($rfds,fileno($file),1)=1;
  }
  local $notdone=1;
  local $/;
  while ($notdone) {
    local $readable;
    local $efds;
    local $nfound=select($readable=$rfds,undef,$efds=$rfds,undef);
    local $x=0;
    while ($x<=$#_) {
      if (vec($readable,fileno($_[$x]),1)) {
        sysread($_[$x],$read,1024);
        $ret[$x].=$read;
      }
      if (vec($efds,fileno($_[$x]),1)) {
        &UI::stderr("slurp_until_eof: Exception on filehandle $_[$x]\n");
      }
      $x++;
    }
  }
}

sub decrypt_message()
{
  # arg0: Reference to message.  {raw} must be defined.  Fills in
  # {userheaders}, {body}, and {decrypterrors}
  local *WFH;
	local *EFH;
	local *RFH;
	if (!open(WFH,">$clientdir/decrypt_temp.$$")) {
	  $_[0]{decrypterrors}.="Could not open tempfile.\n";
	  &UI::stderr("Could not open tempfile.\n");
		return;
	}
	if (!flock(WFH,&LOCK_EX)) {
	  $_[0]{decrypterrors}.="Could not lock tempfile.\n";
		&UI::stderr("Could not lock tempfile.\n");
		close(WFH);
		return;
	}
  local $unmunged=$_[0]{raw};
  $unmunged=~s/^ //mg;
  $unmunged=~s/\012$//s;
	if ((print WFH $unmunged) == 0) {
	  &UI::stderr("Unable to write to tempfile.\n");
		$_[0]{decrypterrors}.="Unable to write to tempfile.\n";
		unlink("$clientdir/decrypt_temp.$$");
		close(WFH);
		return;
	}
	close(WFH);
	local $unbrokenpipe=1;
	local $SIG{PIPE}=sub { $unbrokenpipe=0; };
	local $SIG{CHLD}=sub { $unbrokenpipe=0; };
	$pid=open3(\*WFH,\*RFH,\*EFH,
		'gpg',
		'--secret-keyring', "${clientdir}secring",
		'--keyring', "${clientdir}pubring",
		@extra_gpg_options, #'--quiet',
		'--decrypt', "${clientdir}decrypt_temp.$$");
	local $message='';
	local $notdone=1;
	while ($unbrokenpipe) {
    sysread(RFH,$message,1024,length($message));
    waitpid($pid,&WNOHANG);
  }
  waitpid($pid,0);
  local $oldslash=$/; undef($/);
  $message.=<RFH>;
  close(RFH);
	unlink("$clientdir/decrypt_temp.$$");
	$_[0]{decrypterrors}=<EFH>;
	$/=$oldslash;
	local $headers;
	if (length($message)) {
		if ($message=~/\012\012/s) {
		  ($headers,$_[0]{body})=split(/\012\012/s,$message,2);
		} else {
		  $headers=$message;
			$_[0]{decrypterrors}.="Error: Message contains no body\n";
#			&UI::stderr("Error: Message ($message) contains no body.\n");
		}
	} else {
		$_[0]{unencrypted}=1;
		if ($unmunged =~/\012\012/s) {
		  ($headers,$_[0]{body})=split(/\012\012/s,$unmunged,2);
		} else {
      $_[0]{decrypterrors} .=
        "Error: Message not decryptable; message contains no body.\n";
      &UI::stderr("Error: Message not decryptable; message contains no body.\n");
		}
	}
	local %headers=&split_headers($headers);
	$_[0]{userheaders}=\%headers;
  if (!exists($_[0]{unencrypted})) {
    if (exists($headers{From}) && ($_[0]{decrypterrors}=~/(^|\n)gpg: encrypted with ([^\n]+)\n( +)"([^\n]+)~([^\n]+) \(from ([^\n]+)\)"(\n|$)/s)) {
      if ($6 eq $contacts_recvfrom{$headers{From}}) {
        $_[0]{valid}=1;
        $_[0]{decrypterrors}=~s/(^|\n)gpg: encrypted with ([^\n]+)\n( +)"([^\n]+)~([^\n]+) \(from ([^\n]+)\)"(\n|$)//s;
      } else {
        &UI::stderr("Possible spoofing.\n");
      }
    } else {
      &UI::stderr("Unable to parse GPG output ($_[0]{decrypterrors}).  Falling back to old (slow) &check_from.\n");
      $_[0]{valid} = &check_from($headers{From}, $unmunged, $headers."\012\012".${$_[0]}{body});
    }
  }
}

#sub decrypt_message_orig()
#{
#  local *WFH;
#  local *RFH;
#  local *EFH;
#  local *SPW;
#  local *SPR;
#  pipe(SPR,SPW);
#  $pid=open3(\*WFH,\*RFH,\*EFH,
#    'gpg','--decrypt','--secret-keyring',
#    "$clientdir/secring", '--keyring', "$clientdir/pubring",
#    @extra_gpg_options, '--quiet');
#  local $unmunged=$_[0]{raw};
#  $unmunged=~s/^ //mg;
#  $unmunged=~s/\012$//s;
#  print WFH $unmunged;
#  close(WFH);
#  local $/;
#  local $decrypted=<RFH>;
#  close(RFH);
#  local $headers;
#  local $body;
#  if (length($decrypted)) {
#    ($headers,$body)=split(/\012\012/, $decrypted, 2);
#  } else {
#    ($headers,$body)=split(/\012\012/, $unmunged, 2);
#    $_[0]{unencrypted}=1;
#  }
#  ${$_[0]}{body}=$body;
#  local %headers=&split_headers($headers);
#  $_[0]{userheaders}=\%headers;
#  ${$_[0]}{decrypterrors}=<EFH>;
#  close(EFH);
#  if (!exists($_[0]{unencrypted})) {
#    if (exists($headers{From}) && ($_[0]{decrypterrors}=~/(^|\n)gpg: encrypted with ([^\n]+)\n( +)"([^\n]+)~([^\n]+) \(from ([^\n]+)\)"(\n|$)/s)) {
#      if ($6 eq $contacts_recvfrom{$headers{From}}) {
#        $_[0]{valid}=1;
#        $_[0]{decrypterrors}=~s/(^|\n)gpg: encrypted with ([^\n]+)\n( +)"([^\n]+)~([^\n]+) \(from ([^\n]+)\)"(\n|$)//s;
#      } else {
#        &UI::stderr("Possible spoofing.\n");
#      }
#    } else {
#      &UI::stderr("Unable to parse GPG output.  Falling back to old (slow) &check_from.\n");
#      $_[0]{valid} = &check_from($headers{From}, $unmunged, $decrypted);
#    }
#  }
#  waitpid(-1,WNOHANG);
#}

sub split_filespec()
{
  local $who;
  local $file;
  if (defined($_[0]) && ($_[0] ne '')) {
    if ($_[0]=~/^(.*?)(\/.*?)$/) {
      $who=$1; $file=$2;
    } elsif ($_[0]=~/^(.*?) (.*?)$/) {
      $who=$1; $file=$2;
    } elsif ($_[0]=~/\//) {
      $file=$_[0];
    } elsif ($_[0]=~/~/) {
      $who=$_[0];
    }
  }
  if (defined($who) && ($who!~/~/)) {
    $who=&shortname_to_address($who);
  }
  if (defined($file) && ($file!~/^\//)) {
    $file='/'.$file;
  }
  return ($who,$file);
}

sub split_headers()
{
  local %headers;
  local $key;
  local $val;
  local $x=0;
  local $tmpmsghead=$_[0];
  if (defined($tmpmsghead)) {
    while ($tmpmsghead=~/^(.*?): (.*)$/m) {
      $key=$1; $val=$2;
      $headers{$key}=$val;
      $tmpmsghead=~s/^\Q$key: $val\E$//ms;
      $x++;
      if ($x>1000) {
        &UI::stderr("Infinite loop in client header parsing.  Breaking out.\n");
        $tmpmsghead='';
        &UI::stderr(&clean("Buggy headers: '$_[0]'.\n"));
      }
    }
  }
  return %headers;
}

sub check_from
{
  # Args:
  # 0: From
  # 1: Encrypted Original
  # 2: Decrypted message
  local $retval=0;
  local *WFH;
  local *RFH;
  local *EFH;
  # $_[0]: Claimed sender.
  # $_[1]: Crypted message.
  if (!defined($_[0])) {
    &UI::stderr("Warning: &check_from() sans from.\n");
    return 0;
  }
  if (!exists($contacts_recvfrom{$_[0]})) {
    &UI::stderr("Warning: \$contacts_recvfrom{$_[0]} doesn't exist.\n");
    return 0;
  }
  unlink("$clientdir/temp-check-from");
  system('gpg','--no-default-keyring','--secret-keyring',
    "$clientdir/secring", '--batch', @extra_gpg_options,
    '--keyring', "$clientdir/pubring",
    '--output', "$clientdir/temp-check-from",
    '--export-secret-keys',
    "$Yeemp::Config::my_address (from $contacts_recvfrom{$_[0]})");
  if ($? == 0) {
    open3(\*WFH,\*RFH,\*EFH,'gpg', '--no-default-keyring',
      '--secret-keyring', "$clientdir/temp-check-from",
    '--keyring',"$clientdir/pubring",
      @extra_gpg_options,  '--decrypt');
    local $orig=$_[1];
    $orig=~s/\012\012$//s;
    print WFH "$orig";
    close(WFH);
    local $/;
    local $mess=<RFH>;
    close(RFH);
    local $err=<EFH>;
    close(EFH);
    waitpid(-1,WNOHANG);
    if ($err ne '') {
      UI::stderr("GPG errors $err\n");
    }
    if (!defined($mess)) {
      UI::stderr("Mess undef.\n");
    } elsif (!defined($_[2])) {
      UI::stderr("\$_[2] undef.\n");
    } elsif ($_[2] eq $mess) {
      $retval=1;
    } else {
      &UI::stderr("Consistency error.  May indicate spoofing.\n");
      &clean_print("$_[1]::::$_[2]::::$mess\n");
    }
  } else {
    &UI::stderr("Unable to export a key for '$Yeemp::Config::my_address (from $contacts_recvfrom{$_[0]})'.\n");
  }
  return $retval;
}


sub handle_statchange()
{
  if ($_[2]=~/^Received: ([0-9]+)$/m) {
    if ($1>$contacts_status_updated{$_[0]}) {
      $contacts_status_updated{$_[0]}=$1;
      if ($contacts_status{$_[0]} ne $_[1]) {
        $contacts_status{$_[0]}=$_[1];
        &UI::show_statchange($_[0],$_[1]);
        play_statchange($_[0],$_[1]);
      }
    }
  } else {
    &UI::stderr("Malformed status-change (no Received: server-header) received.\n");
  }
  if ($Yeemp::Config::cache_status) {
    &cache($_[0],'/status',"$_[2]\012\012 From: $_[0]\012Status: $_[1] [Cached]\012\012");
  }
}

sub sanity_check_pubkey()
{
  # Arg0: User
  # Arg1: Key
  local $sanity=0;
  local $warnings='';
  local $current_key=&extract_sending_public_key($_[0]);
  if (defined($current_key) && ($current_key ne '')) {
    if ($current_key ne $_[1]) {
      $warnings.=
        "WARNING:  This key differs from the one you already have for\n".
        "$_[0].  This may mean that someone is trying to\n".
        "impersonate $_[0], or that someone already is.\n";
      $sanity--;
    } else {
      $warnings.="Warning: Duplicate key received from ".$_[0]."\n";
    }
  }
  &UI::stderr("DEBUG: Add code to list prior fingerprints received.\n");
  return ($sanity,$warnings);
}

sub save_file_as()
{
	local *FILE;
	if ($_[0]=~/^(.*?)$/s) {
		$_[0]=$1;
	} else {
		&UI::stderr("Save cancelled.\n");
	}
	if (sysopen(FILE,"$_[0]",O_EXCL|O_CREAT|O_WRONLY)) {
		if (flock(FILE,LOCK_EX)) {
			print FILE $_[1];
			flock(FILE,LOCK_UN);
		} else {
			&UI::stderr("Unable to lock file $_[0]\n");
		}
		close(FILE);
	} else {
		&UI::stderr("Unable to write file $_[0]\n");
	}
}

sub save_unknown_file()
{
  local $saveas=${$_[0]}[1];
	if ($saveas=~/\/([^\/]+)$/s) {
		$saveas=$1;
	}
	if ($saveas!~/^([-0-9A-Za-z_.]+)$/s) {
		$saveas=~s/[^-0-9A-Za-z_.]/_/gs;
	}
	&UI::read_line("Save as: ",'Save',\&save_file_as, $saveas, ${$_[0]}[2])
}

sub handle_unknown_file()
{
	&UI::ynprompt("File ".&clean($_[1])." received from $_[0].\nSave it?",
		\&save_unknown_file,\@_)
}

sub validate_portrait()
{
	local ($address, $filename) = @_;
	if ($filename=~/\.(png|jpg)$/s) {
		unlink("$clientdir/cache/$address/portrait.xpm");
		local $ext=$1;
		if ($ext eq 'jpg') {
			unlink("$clientdir/cache/$address/portrait.png");
			$ext = 'jpeg';
		} else {
			unlink("$clientdir/cache/$address/portrait.jpg");
		}
		local @ppm = split /[ \t\r\n]+/,`${ext}topnm $clientdir/cache/$address$filename`;
		if ($ppm[1]<=$Yeemp::Config::maxportraitsize &&
			$ppm[2]<=$Yeemp::Config::maxportraitsize) {
			return 1;
		} else {
			&UI::ynprompt(
				"Oversize portrait image received from $address
				($ppm[1] x $ppm[0] pixels).  Use it anyhow?",
				sub { }, undef, sub { unlink("$clientdir/cache/$address$filename"); });
			return 1;
		}
	} else {
		&UI::stderr("Invalid portrait filename $filename.\n");
	}
	sleep(5);
	return 0;
}

sub handle_file()
{
  # Args: 0: Ref to message hash.
  local $file=$_[0]{userheaders}{File};
  local $from=$_[0]{userheaders}{From};
  if ($file eq '/public_key') {
    local $key=$_[0]{body};
    &import_temp_pubkey($from,$key);
    local @fingerprint=&get_tempkey_fingerprints();
    local ($sanity,$warnings)=&sanity_check_pubkey($from,$key);
    if ($warnings ne '') {
      &UI::stderr("Warnings: $warnings");
    }
    &UI::show_inbound_public_key($from,$key);
  } elsif ($file eq '/status') {
    &handle_statchange($from, $_[0]{body}, $_[0]{serverheaders});
  } elsif ($file =~ /^\/statsounds\/(.*?)\.ogg$/s) {
    my $statsound = $1;
    if ($statsound!~/..\//s) {
      cache($from, $file, $_[0]{body});
    }
	} elsif ($file eq '/beep.ogg') {
	  &cache($from, '/beep.ogg', $_[0]{body});
	} elsif ($file =~ /^\/portrait\.(png|jpg)$/s) {
	  &cache($from, $file, $_[0]{body});
		if (&validate_portrait($from, $file)) {
			&UI::show_image($from, "$from has changed their portrait.", $file);
		}
	} elsif ($file =~ /^\/snapshot\.(png|jpg)$/s) {
	  &cache($from, $file, $_[0]{body});
		&UI::show_image($from, "\%s has sent you a snapshot.", $file);
	} else {
	  &handle_unknown_file($from, $file, $_[0]{body});
	}
}

sub send_user_file()
{
	local %args=%{$_[0]};
	if (-f $args{sendas}) {
	  $args{file}=$args{sendas};
		&send_arbitrary_file(\%args);
	} else {
	  &UI::read_line("$args{sendas} doesn't exist.\nFile to send as $args{sendas}: ",
		  'OK', sub {
			  local %args=%{$_[1]}; 
				if (-f $_[0]) {
				  $args{file}=$_[0];
					&send_arbitrary_file(\%args);
				} else {
				  &UI::stderr("Unable to read file $_[0].\n");
				}
			}, '', \%args );
	}
}

sub handle_get($)
{
  # $_[0]: Ref to message hash
  local $file = $_[0]{userheaders}{Get};
	local $from = $_[0]{userheaders}{From};
  if ($file eq '/public_key') {
    local $key;
    if ($_[0]{body} ne '') {
      $key=$_[0]{body};
      $key=~s/\012$//s;
    } else {
      $key=$_[0]{raw};
      $key=~s/^(.*?)\012 //s;
      $key=~s/^ //mg;
      $key=~s/\012\012$//s;
    }
    &import_temp_pubkey($from,$key);
    local @print=&get_tempkey_fingerprints();
    if (!defined($print[0])) {
      $print[0]='[Internal error obtaining key fingerprint]';
    }
    &UI::show_public_key_request($_[0]{userheaders}{From},
      $key,$print[0]);
    return 1;
  }
  if ($_[0]{unencrypted}) {
    UI::stderr("Warning: Unencrypted Get request received.\n");
    return 0;
  }
  if (!$_[0]{valid}) {
    UI::stderr("Warning: Apparent spoofed file request.  Ignoring.\n");
    return 0;
  }
  if ($file eq '/status') {
#    local $from=$_[0]{userheaders}{From};
    if (exists($contacts_presence{$from})) {
      &handle_statreq($from, $contacts_presence{$from});
    } else {
      &handle_statreq($from, $Yeemp::Config::default_presence_mode);
    }
  } else {
		local %args=('to'=>$from,'sendas'=>$file);
		&UI::ynprompt("$from has requested $file.  Send it?",
			\&send_user_file, \%args);
  }
}

sub handle_statreq()
{
  local $from=$_[0];
  local $mode=$_[1];
  if (!exists($contacts_sendfrom{$from})) {
    &UI::stderr("Status request from $from, whose public key you don't have.  Skipping.\n");
    return;
  }
  if ($mode eq 'confirm') {
    &UI::ynprompt("Notify $from of your current status? ", \&send_status,
      $from);
  } elsif ($mode eq 'custom') {
    &UI::stderr("Custom status not yet implemented.\n");
  } elsif ($mode eq 'hide') {
    &UI::stdout("Hiding from $from.\n");
  } elsif ($mode eq 'show') {
    &deliver_message($from,"From: $Yeemp::Config::my_address\012Status: $my_status\012\012", $Yeemp::Config::status_change_expiry) or &UI::stderr("Error delivering status to $from.");
  } else {
    &UI::stderr("Status request, and you have no default status policy.\n");
  }
}

sub send_status()
{
  if (!defined($my_status)) {
    print STDERR "Status undefined.\n";
  }
  &deliver_message($_[0],
    "From: $Yeemp::Config::my_address\012Status: $my_status\012",
    $Yeemp::Config::status_change_expiry) or
      &UI::stderr("Error delivering status to $from.");
}

sub handle_message()
{
  local %message;
  local %headers;
  if ($_[0]=~/^([^ ].*?)\012( .*?)$/s) {
    $message{serverheaders}=$1;
    $message{raw}=$2;
    local %serverheaders=&split_headers($message{serverheaders});
    &decrypt_message(\%message);
    %headers=%{$message{userheaders}};
    $last_raw=$_[0];
    if (exists($headers{From})) {
      if ($message{valid}) {
        &log_message($headers{From},$message{serverheaders}.$message{raw});
        $last_from=$headers{From};
        $last_subfrom=$headers{Subfrom};
      } else {
        &log_message("Unvalidated", $message{serverheaders}.$message{raw});
      }
    } else {
      &log_message("No_From", $message{serverheaders}.$message{raw});
    }
    if (exists($headers{File})) {
      &handle_file(\%message);
    }
    my $handled = 0;
    if (exists($headers{Get})) {
      $handled = handle_get(\%message);
    }
    if (exists($headers{Status}) && $message{valid}) {
      &handle_statchange($headers{From}, $headers{Status}, $message{serverheaders});
    }
    if ((!$handled) && (!exists($headers{File}))) {
      if (exists($message{body}) && ($message{body} ne '')) {
        &UI::show_message(\%message);
			}
    }
    local @resp=channel_command(CHANNEL,"Delete: $serverheaders{'Message ID'}");    if ($resp[0] ne '1') {
      &UI::stderr("Error deleting logged message '$serverheaders{'Message ID'}'\n(@resp)\n");
    }
  } else {
	  $message{serverheaders}=$_[0];
    local %serverheaders=&split_headers($message{serverheaders});
		&UI::show_corrupt_message($_[0]);
		local @resp=
			channel_command(CHANNEL,"Delete: $serverheaders{'Message ID'}");
		if ($resp[0] ne '1') {
			&UI::stderr("Error deleting message '$serverheaders{'Message ID'}'\n");
		}
    &UI::stderr("Regex mismatch.\n");
  }
  return 1;
}

sub import_imbox()
{
  local *DEST;
  local *SRC;
  local $ret=0;
  if (!(-f $_[1])) {
    # If the file doesn't exist, no need to try.
    return;
  }
  if (open (SRC,"+<$_[1]")) {
    if (flock(SRC,LOCK_EX|LOCK_NB)) {
      if (open(DEST,">>$_[0]")) {
        if (flock(DEST,LOCK_EX|LOCK_NB)) {
          seek(DEST,0,2);
          local($/);
          local $src=<SRC>;
          if (length($src)>0) {
            $ret=(print DEST $src);
          }
          truncate(SRC,0);
          flock(DEST,LOCK_UN);
        } else {
          &UI::stderr("Error locking destination file $_[0].\n");
        }
      } else {
        &UI::stderr("Error opening destination file $_[0].\n");
      }
      flock(SRC,LOCK_UN);
    } else {
      &UI::stderr("Error: Cannot lock file $_[1].\n");
    }
    close(SRC);
  } else {
    &UI::stderr("Error: Cannot read messages from '$_[1]'.\n");
  }
  return $ret;
}

sub file_snip
{
  local $/;
  local *FILE=$_[0];
  local $start=$_[1];
  local $end=$_[2];
  seek(FILE,0,SEEK_SET);
  local $file=<FILE>;
  substr($file,$start,$end-$start)='';
  seek(FILE,0,SEEK_SET);
  print FILE "$file";
  local $filelen=length($file);
  if ($filelen=~/([0-9]+)/) {
    $filelen=$1;
  }
  truncate(FILE,tell(FILE));
  seek(FILE,$start,SEEK_SET);
}

sub keyring_add()
{
  # Arg0: Keyring
  # Arg1: Key
  local *WFH;
  local *RFH;
  local *EFH;
  local $ret=0;
  open3(\*WFH, \*RFH, \*EFH, 'gpg', '--no-default-keyring',
    '--keyring', $_[0], '--import', '--allow-non-selfsigned-uid',
    '--status-fd', 1, @Yeemp::Client::extra_gpg_options);
  print WFH $_[1];
  close(WFH);
  local $line;
  local @lines;
  while ($line=<RFH>) {
    push(@lines,$line);
  }
  local $/;
  close(RFH);
  local $err=<EFH>;
  if (defined($err) && ($err ne '')) {
    &UI::stderr("Debug: \&keyring_add(): GPG errors:\n$err\n");
  }
  waitpid(-1,WNOHANG);
  return @lines;
}

sub lock_filename()
{
  local *LOCK;
  if (sysopen(LOCK,"$_[0].yeemplock",O_CREAT|O_EXCL)) {
    close(LOCK);
    return 1;
  }
  return 0;
}

sub unlock_filename()
{
  unlink("$_[0].yeemplock");
}

sub get_tempkey_fingerprints()
{
  # returns a list: Fingerprint, Yeemp address, Yeemp fromtag
  local @ret=undef;
  if (&lock_filename("$clientdir/temp-keyring")) {
    local $fingerprints=`gpg --list-keys --with-colons --no-default-keyring --keyring $clientdir/temp-keyring --fingerprint`;
    &UI::debug("Fingerprints: '$fingerprints'\n");
    $fingerprints=~s/^(.*?)--\n//s;
    local @prints=split("\n",$fingerprints);
    local @line=split(':',$prints[0]);
    if ($line[0] eq 'pub') {
      if ($line[9]=~/^(.*?~.*?) \(from (.*?)\)$/) {
        $ret[1]=$1;
        $ret[2]=$2;
      }
    }
    @line=split(':',$prints[1]);
    if ($line[0] eq 'fpr') {
      $ret[0]=$line[9];
    }
    &unlock_filename("$clientdir/temp-keyring");
  }
  return @ret;
}

sub note_tempkey_fingerprints()
{
  local *FILE;
  local $fingerprints=
    `gpg --list-keys --with-colons --no-default-keyring --keyring $clientdir/temp-keyring --fingerprint`;
  $fingerprints=~s/^(.*?)-\n//s;
  local @lines=split("\n",$fingerprints);
  if (open(FILE,">>$clientdir/fingerprints")) {
    local $ret=1;
    if (flock(FILE,LOCK_EX)) {
      if (seek(FILE,0,SEEK_END)) {
        $ret &= print FILE "-- ".localtime(time())." --\n";
        $ret &= print FILE $fingerprints;
      } else {
        $ret=0;
      }
      flock(FILE,LOCK_UN);
    }
    close(FILE);
    return $ret;
  }
  return 0;
}

sub import_temp_pubkey()
{
  # Arg0: Address
  # Arg1: Key
  if (&lock_filename("$clientdir/temp-keyring")) {
    unlink("$clientdir/temp-keyring");
    local @lines=&keyring_add($clientdir.'temp-keyring',$_[1]);
    local $ret=&note_tempkey_fingerprints();
    &unlock_filename("$clientdir/temp-keyring");
    return $ret;
  }
  &UI::stderr("Unable to create $clientdir/temp-keyring.yeemplock.\nQuit all Yeemps, delete it manually, and try again.\n");
  return 0;
}

sub import_contact_pubkey()
{
  # Arg0: Address
  # Arg1: Key
  &UI::debug("Import_contact_pubkey()\n");
  local @lines = &keyring_add( $clientdir.'pubring' ,$_[1]);
  @lines=grep {/^\[GNUPG:\] IMPORTED/} @lines;
  if (@lines && ($lines[0]=~/^\[GNUPG:\] IMPORTED ([A-Z0-9]+) (.*?)$/)) {
    local $keyid=$2;
    if ($keyid=~/^(.*?) \(from (.*?)\)$/) {
      local $address=$1;
      local $sendfrom=$2;
      if ($address ne $_[0]) {
        &UI::stderr("Key address and expected address don't match.\n");
        return 0;
      }
      if (&set_contact_sendfrom($address,$sendfrom)) {
        $ret=1;
      }
    } else {
      &UI::stderr("Error: keyid syntax problems, confused client.\n");
    }
  } else {
    &import_temp_pubkey($_[0],$_[1]);
    local $key=`gpg --list-keys --with-colons --no-default-keyring --keyring ${clientdir}temp-keyring $_[0]`;
    if ($key=~/^pub:.*?:.*?:.*?:.*?:.*?:.*?:.*?:.*?:(.*?):.*?$/m) {
      local $to=$1;
      if ($to=~/\Q$_[0] (\Efrom (.*?)\)$/) {
        local $sendfrom=$1;
        &set_contact_sendfrom($_[0],$sendfrom);
        return 1;
      }
    }
    &UI::stderr("Error importing public key - unable to determine origin.\n");
  }
  if ($ret) {
    resend_bare_postponed_spool();
  }
  return $ret;
}

#sub import_public_key()
#{
#  @lines=grep {/^\[GNUPG:\] IMPORTED/} @lines;
#  if (@lines && ($lines[0]=~/^\[GNUPG:\] IMPORTED ([A-Z0-9]+) (.*?)$/)) {
#    local $keyid=$2;
#    if ($keyid=~/^(.*?) \(from (.*?)\)$/) {
#      local $address=$1;
#      local $sendfrom=$2;
#      if (&set_contact_sendfrom($address,$sendfrom)) {
#        $ret=1;
#      }
#    } else {
#      print "Warning: keyid syntax problems, confused client.\n";
#    }
#  }
#  print <EFH>;
#  waitpid(-1,WNOHANG);
#  return $ret;
#}


sub grant_public_key_request()
{
  # $_[0]: Requestor's address
  # $_[1]: Requestor's public key.
  local *RFH;
  local *WFH;
  local *EFH;
  local @lines;
  if (&import_temp_pubkey($_[0],$_[1])) {
    local $pubkey=&create_recipient_public_key($_[0]);
    &UI::debug("Recipient public key is '$_[0]'\n");
    local $keydata=`gpg --list-keys --with-colons --no-default-keyring --keyring $clientdir/temp-keyring`;
    local %contacts_sendfrom;
    if ($keydata=~/^(.*?):(.*?):(.*?):(.*?):(.*?):(.*?):(.*?):(.*?):(.*?):(.*?):(.*?)$/m) {
      local $keydata=$10;
      if ($keydata=~/^\Q$_[0] (from \E(.*?)\)$/) {
         $contacts_sendfrom{$_[0]}=$1;
      } else {
        &UI::stderr("Warning: Could not find sendfrom field for $_[0] in $keydata.\n");
      }
    } else {
      &UI::stderr("Warning: Could not find from for $_[0] in $keydata.\n");
      return 0;
    }
    local $mess=&encrypt_message_to_tempkey($_[0],
      "From: $Yeemp::Config::my_address\012File: /public_key\012\012$pubkey");
    if ((!defined($mess)) || ($mess eq '')) {
      &UI::stderr("Warning: Error encrypting message.\n");
      return 0;
    }
    &lazy_open_channel();
    if (&send_message($_[0],$mess)==0) {
      &spool_postponed_message_crypted(@_);
    }
    return 1;
  }
  return 0;
}

sub extract_sending_public_key()
{
  # Extracts a key to be used for sending messages.
  local $to=$_[0];
  if (!defined($contacts_sendfrom{$_[0]})) {
    &UI::stdout(
      "Warning: contacts_sendfrom{$_[0]} undefined.\n");
    return undef;
  }
  if (exists($contacts_sendfrom{$_[0]})) {
    return &extract_public_key("$_[0] (from $contacts_sendfrom{$_[0]})");
  }
  return undef;
}

sub unmunge_stored()
{
  local $unmunged=$_[0];
  $unmunged=~s/^ //mg;
  $unmunged=~s/\012$//s;
  return $unmunged;
}

sub update_status($)
{
  $my_status=$_[0];
  if (!defined($Yeemp::Config::my_address)) {
    &UI::stderr("my_address not defined.\n");
  }
  $message="From: $Yeemp::Config::my_address\012File: /status\012\012$my_status";
  local $cont;
  # XXX deal with contacts whose presence is set to 'confirm' better.
  foreach $cont (keys(%contacts)) {
    my $auth =
      ($contacts_presence{$_[0]} || $Yeemp::Config::default_presence_mode);
    if ($auth eq 'show') {
      &deliver_message($cont,$message,$Yeemp::Config::status_change_expiry);
    } elsif ($auth eq 'confirm') {
      UI::ynprompt("Notify $cont of your current status? ", \&send_status,
        $cont);
    }
  }
}

sub announce_status()
{
  $my_status=$Yeemp::Config::default_status;
  if ($Yeemp::Config::announce_status) {
    if (!defined($Yeemp::Config::my_address)) {
      &UI::stderr("my_address not defined.\n");
    }
    $message="From: $Yeemp::Config::my_address\012File: /status\012Get: /status\012\012$my_status";
    local $cont;
    foreach $cont (keys(%contacts)) {
      local $show=0;
      if (exists($contacts_presence{$cont})) {
        if ($contacts_presence{$cont} eq 'show') {
          $show=1;
        }
      } elsif ($Yeemp::Config::default_presence_mode eq 'show') {
        $show=1;
      }
      if ($show) {
        &deliver_message($cont,$message,
          $Yeemp::Config::status_change_expiry);
      }
    }
  }
}

sub file_is_blank($)
{
  if (!(-f $_[0])) {
    &UI::stderr("Debug: File $_[0] doesn't exist.\n");
    return 1;
  }
  local @stat=stat($_[0]);
  if ($stat[7]==0) {
    return 1;
  }
  return 0;
}

sub generate_login_key()
{
  local (*WFH, *RFH, *EFH, *STATR, *STATW);
  # Arg 0: Server name
  # Arg 1: Address
  my $oldctrlf = $^F;
  $^F=1023;
  if (!pipe(STATR, STATW)) {
    $^F=$oldctrlf;
    UI::stderr("Pipe failed.\n");
    return;
  }
  $^F=$oldctrlf;
  &UI::debug("Generating login key: $_[1]\n");
  &UI::stdout("Generating private key.  This may take several minutes.\n");
  open3(\*WFH,\*RFH,\*EFH,'gpg','--no-default-keyring',
    '--status-fd', fileno(STATW),
    '--secret-keyring', "${clientdir}login-key-$_[1]",
    '--keyring', "${clientdir}login-key-$_[1].pub",
    '--gen-key','--batch', @extra_gpg_options);
  print WFH "Key-Type: DSA\n";
  print WFH "Key-Length: 1024\n";
  print WFH "Subkey-Type: ELG-E\n";
  print WFH "Subkey-Length: 1024\n";
  print WFH "Name-Real: $_[1]\n";
  print WFH "Name-Comment: login key for $_[0]\n";
  print WFH "%commit\n";
  close(WFH);
  local $fds='';
  vec ($fds, fileno(RFH), 1)=1;
  vec ($fds, fileno(EFH), 1)=1;
  vec ($fds, fileno(STATR), 1)=1;
  local $readable;
  local $nfound;
  local $skipout=0;
  local $skiperr=0;
  local $errblock='';
  local $outblock='';
  my $key_created=0;
  while ($nfound=select($readable=$fds, undef, undef, undef)) {
    if (vec($readable,fileno(EFH),1)) {
      local $err;
      local $read=sysread(EFH,$err,128);
      if ($read==0) {
        vec($fds, fileno(EFH), 1) = 0;
        $skiperr=1;
      }
      $errblock.=$err;
      if ($errblock=~/^(.*)\n/s) {
        local $err=$1;
        $errblock=~s/^\Q$err\E\n//s;
        &UI::stderr("Error generating login key: $err\n");
      }
    }
    if (vec($readable,fileno(RFH),1)) {
      local $out;
      local $read=sysread(RFH, $out, 128);
      if ($read==0) {
        vec($fds, fileno(RFH), 1) =0;
        $skipout=1;
      }
      $outblock.=$out;
      if ($outblock=~/^(.*)\n/s) {
        local $out=$1;
        $outblock=~s/^\Q$out\E\n//s;
        &UI::stderr("Unexpected output generating login key: $out\n");
      }
    }
    if (vec($readable, fileno(STATR), 1)) {
      my $out;
      my $read = <STATR>;
      if ($read=~/^\[GNUPG:\] PROGRESS (.*?)$/) {
        # XXX update progress widget, which doesn't yet exist
      } elsif ($read=~/^\[GNUPG:\] GOOD_PASSPHRASE$/) {
        # Nothing.
      } elsif ($read=~/^\[GNUPG:\] KEY_CREATED B$/) {
        $key_created=1;
      } else {
        UI::stderr("Got unexpected status output $read");
      }
    }
    if ($skipout && $skiperr) {
      last;
    }
  }
  if (!$key_created) {
    UI::stderr("Key not created.\n");
  }
  if ($outblock ne '') {
    UI::stderr("Unexpected output generating login key: $outblock\n");
  }
  if ($errblock ne '') {
    UI::stderr("Errors generating login key: $errblock\n");
  }
  close(RFH);
  close(EFH);
#  do {
#    $kid = waitpid(-1,&WNOHANG);
#    &UI::debug("Kidpid: $kid\n");
#  } until $kid == -1;
}


sub copy_login_key()
{
  &UI::ynprompt("Is $_[0] the machine you're running Yeemp on? ",
    \&callback_copy_login_key_local, \@_, \&callback_copy_login_key_remote);
}

sub callback_copy_login_key_remote()
{
  @_=@{$_[0]};
  &UI::ynprompt("Copy the login key to $_[0] using SCP? ",\&callback_copy_login_key,
    $_[1], \&callback_copy_login_key_no);
}

sub callback_copy_login_key_local()
{
  local @pwent=getpwuid($<);
  local $umask=umask(022);
  local *FILE;
  if (open(FILE,"<${clientdir}login-key-$Yeemp::Config::my_address.pub")) {
    local $/;
    local $file=<FILE>;
    close(FILE);
    if (sysopen(FILE,"$pwent[7]/.yeemp-login-key-$Yeemp::Config::my_address.pub",O_CREAT|$Yeemp::Config::O_NOFOLLOW|O_WRONLY)) {
      print FILE $file;
      close(FILE);
    } else {
      &UI::stderr("Error opening public key for writing.\n");
    }
  } else {
    &UI::stderr("Error opening public key for reading.\n");
  }
  reconnect_to_server();
}

sub callback_copy_login_key()
{
  local ($user,$machine)=split('~',$_[0]);
  &UI::read_line("Account to copy the login key to: ", "OK",
    \&callback_copy_login_key_to_acct, "$user\@$machine");
}

sub scp_complete()
{
  reconnect_to_server();
}

sub callback_copy_login_key_to_acct()
{
  if (!eval("use Yeemp::SCP; 1;")) {
    UI::stderr("Error: Could not load Yeemp::SCP;\n");
    return;
  }
  local $scp = new Yeemp::SCP( 
    'sourcefile', "${clientdir}login-key-$Yeemp::Config::my_address.pub",
    'destacct', $_[0],
    'destfile', ".yeemp-login-key-$Yeemp::Config::my_address.pub",
    'callback', \&scp_complete );
  $scp->cycle();
}

sub parse_args(@)
{
  local %args;
  while (@_) {
    local $arg=shift;
    if ($arg eq '-d') {
      $arg=shift;
      if (defined($arg)) {
        if ($arg=~ /^([-\/0-9_.a-zA-Z]+)$/s) {
          $arg=$1;
          if (!(-d $arg)) {
            if (mkdir($arg,0700)) {
              &UI::stdout("Created $arg\n");
            } else {
              &UI::stdout("Unable to create $arg\n");
            }
          }
          if (-d $arg) {
            &UI::stdout("Using $arg as Yeemp config dir.\n");
            $args{clientdir}=$arg;
          }
        } else {
          &UI::stderr("Unable to untaint $arg\n");
        }
      } else {
        &UI::stderr("-d requires an argument.\n");
      }
    } elsif ($arg eq '-f') {
      $Yeemp::Config::fast_start=1;
    } elsif ($arg eq '-v') {
      $Yeemp::Config::debug=1;
    } else {
      &UI::stderr("Syntax: yeemp [-d clientdir] [-v] [-f]\nxeemp [-d clientdir] [-v] [-f]\n");
    }
  }
  return %args;
}

sub store_valid_cert()
{
  # Arg: hash of domain, cert
  local %args=%{$_[0]};
  &UI::debug("store_valid_cert ($Yeemp::Config::certcache_dir/$args{domain})\n");
  local *PEM;
  if (open (PEM,">$Yeemp::Config::certcache_dir/$args{domain}")) {
    if (flock(PEM,LOCK_EX)) {
      if ((print PEM $args{cert})) {
        $Yeemp::Config::rekeyed{$args{domain}}=1;
      } else {
        &UI::stderr("Error writing cert.\n");
      }
      flock(PEM,LOCK_UN);
    } else {
      &UI::stderr("Error locking $Yeemp::Config::certcache_dir/$args{domain}\n");
    }
    close(PEM);
  } else {
    &UI::stderr("Error opening $Yeemp::Config::certcache_dir/$args{domain}\n");
  }
}

sub validate_unknown_cert()
{
  # Arg0: Domain
  # Arg1: PEM-encoded cert
  local $domain=$_[0];
  if ($domain=~/^([-_a-zA-Z.]+)$/) {
    $domain=$1;
  } else {
    &UI::stderr("Error untainting domain name $_[0]\n");
  }
  local *FILE;
  if (open(FILE,">${clientdir}tempcert-$domain")) {
    if (flock(FILE,LOCK_EX)) {
      local $old_fh = select(FILE);
      $| = 1;
      select($old_fh);
      print FILE $_[1];
      local $fingerprint=
        `openssl x509 -in ${clientdir}tempcert-$domain -noout -sha1 -fingerprint`;
      chomp($fingerprint);
      local %hash=(domain=>$domain, cert=>$_[1]);
      &UI::ynprompt(
        "\nAdd cert for $domain to certlist\n($fingerprint)?\n",
        \&Yeemp::Client::store_valid_cert, \%hash);
      flock(FILE,LOCK_UN);
    } else {
      &UI::stderr("Error locking ${clientdir}tempcert-$_[0]");
    }
    unlink("${clientdir}tempcert-$_[0]");
    close(FILE);
  } else {
    &UI::stderr("Unable to write to ${clientdir}tempcert-$_[0]");
  }
}

sub callback_copy_login_key_no()
{
  print "You'll need to copy ${clientdir}login-key-$Yeemp::Config::my_address.pub\n";
  print "to your server and install it as your public login key before you can\n";
  print "login.\n";
}

sub callback_generate_login_key()
{
  &generate_login_key($Yeemp::Config::server_name, $Yeemp::Config::my_address);
  &copy_login_key($Yeemp::Config::server_name, $Yeemp::Config::my_address);
}

sub check_login_key()
{
  if ($Yeemp::Config::client_mode ne 'remote') {
    # Don't care about the login key if it won't be used...
    return 1;
  }
  local $nokey=0;
  if (&file_is_blank("${clientdir}login-key-$Yeemp::Config::my_address.pub")) {
    $nokey=1;
  } elsif (&file_is_blank("${clientdir}login-key-$Yeemp::Config::my_address")) {
    $nokey=1;
  }
  if ($nokey) {
    &UI::ynprompt("Login key for $Yeemp::Config::server_name not found.  Generate it? ",\&callback_generate_login_key, undef,
      \&shutdown_client);
#    $Yeemp::Config::offline{$Yeemp::Config::server_name}=1;
  }
}

sub check_postponed_spools()
{
  &resend_bare_postponed_spool();
  &resend_crypted_postponed_spool();
}

sub extract_receiving_public_key()
{
  # Extract the public component of a key which will be used to send 
  # messages to you.
  if (!defined($Yeemp::Config::my_address)) {
    &UI::stderr("Warning: my_address not defined.\n");
    return undef;
  }
  if (exists($contacts_recvfrom{$_[0]})) {
    return &extract_public_key("$Yeemp::Config::my_address (from $contacts_recvfrom{$_[0]})");
  }
  return undef;
}

sub extract_public_key()
{
  local $keyname=$_[0];
  if ($keyname=~/^([() ~.a-zA-Z0-9_]+)$/) {
    $keyname=$1;
  } else {
    &UI::stderr("Error: Cannot untaint name '$_[0]'\n");
    return;
  }
  local $extraopts=join(' ',@extra_gpg_options);
  local $key=`gpg $extraopts --no-default-keyring --secret-keyring ${clientdir}secring --keyring ${clientdir}pubring --export '$keyname'`;
  return $key;
}

sub new_contacts_recvfrom()
{
  local $k;
  local $notdone=1;
  local $ret;
  while ($notdone) {
    $notdone=0;
    $ret=rand(10000);
    foreach $k (keys(%contacts_recvfrom)) {
      if ($contacts_recvfrom{$k} eq $ret) {
        $notdone=1;
      }
    }
  }
  return $ret;
}

sub generate_keypair
{
  # $_[0]: User that this key will be used to receive messages from
  local *WFH;
  local *RFH;
  local *EFH;
  if (!exists($contacts{$_[0]})) {
    $contacts{$_[0]} = undef();
  }
  if (!exists($contacts_recvfrom{$_[0]})) {
    $contacts_recvfrom{$_[0]}=&new_contacts_recvfrom();
  }
  save_contact_list();
  &UI::stdout("Generating a keypair.  This may take a long time.\n");
  open3(\*WFH,\*RFH,\*EFH,'gpg','--no-default-keyring',
    '--secret-keyring', "${clientdir}secring",
    '--keyring', "${clientdir}pubring",
    '--gen-key','--batch', @extra_gpg_options);
  print WFH "Key-Type: DSA\n";
  print WFH "Key-Length: 1024\n";
  print WFH "Subkey-Type: ELG-E\n";
  print WFH "Subkey-Length: 1024\n";
  print WFH "Name-Real: $Yeemp::Config::my_address\n";
  print WFH "Name-Comment: from $contacts_recvfrom{$_[0]}\n";
  print WFH "%commit\n";
  close(WFH);
#  local $oldslash=$/;
  local($/);
  &UI::debug("<RFH...>\n");
  print <RFH>;
  &UI::debug("<EFH...>\n");
  print <EFH>;
  close(RFH);
  close(EFH);
 # $/=$oldslash;
  waitpid(-1,WNOHANG);
  return &extract_receiving_public_key($_[0]);
}

sub add_contact($$)
{
  # $_[0]: Address
  # $_[1]: Shortname
  if (exists($contacts{$_[0]})) {
    return 0;
  }
  $contacts{$_[0]} = 1;
  if (defined($_[1]) && length($_[1])) {
    $contacts_shortnames{$_[0]}=$_[1];
  }
  if (!exists($contacts_status{$_[0]})) {
    $contacts_status{$_[0]}='[Unknown]';
  }
  save_contact_list();
  return 1;
}

sub create_recipient_public_key
{
  local $key=&extract_receiving_public_key($_[0]);
  if (!defined($key) || ($key eq '')) {
    $key=&generate_keypair($_[0]);
  }
  return $key;
}

sub spool_postponed_message_bare()
{
  &spool_message($Yeemp::Config::postponed_spool_bare,$_[0],$_[1]);
}

sub request_public_key()
{
  if (!defined($Yeemp::Config::my_address)) {
    &UI::stderr("No sender address defined.  Cannot request key.\n");
    return;
  }
  local $message="From: $Yeemp::Config::my_address\012Get: /public_key\012\012".
  &create_recipient_public_key($_[0]);
  if (&lazy_open_channel()) {
    if (&send_message($_[0],$message)==0) {
      &spool_postponed_message_bare($_[0],$message);
    }
  } else {
    &UI::stderr("Error sending public key request.\n");
    &spool_postponed_message_bare($_[0],$message);
  }
}

sub slurp_define()
{
  local *INCLUDE;
  if (open(INCLUDE,"<$_[1]")) {
    local @include=<INCLUDE>;
    @include=grep{/^#[ \t]*define(.*?)\Q$_[0]\E(.*?)([0-9x]+)/} @include;
    if (!@include) {
      return undef;
    }
    if ($include[0]=~/^#[ \t]*define(.*?)\Q$_[0]\E(.*?)([0-9x]+)/) {
    local $val=$3;
    if ($val=~/^0/) {
      return oct($val);
    } else {
      return $val;
    }
    return undef;
    }
  }
  return undef;
}

sub flowinout()
{
  local $readfh=$_[0];
	local $writefh=$_[1];
	local $writedata=$_[2];
	local $readdata='';
	# Make the pipes raw.
	local $lastfile=select($readfh);
	$|=1; select($writefh); $|=1;
	select($lastfile);
	local $wfd='';
	vec($wfd,fileno($writefh),1)=1;
	local $rfd='';
	vec($rfd,fileno($readfh),1)=1;
  local $readable;
	local $writeable;
	while (length($writedata)) {
		select($readable=$rfd, $writeable=$wfd, undef, undef);
		if (vec($wfd,fileno($writefh),1)) {
		  # Writeable
			if (length($writedata)>4096) {
			  $len=4096;
			} else {
			  $len=length($writedata);
			}
			local $wrote=syswrite($writefh,$writedata,$len);
			substr($writedata,0,$wrote)='';
		}
		if (vec($rfd,fileno($readfh),1)) {
			local $num='';
			ioctl($readfh,&FIONREAD, $num);
			$num=unpack('I',$num);
			if ($num) {
				sysread($readfh,$readdata,$num,length($readdata));
			}
		}
	}
	close($writefh);
	while ((select($readable=$rfd, undef, undef, 0.5))[0]) {
	  local $num='';
	  ioctl($readfh,&FIONREAD,$num);
		$num=unpack('I',$num);
		&UI::debug("Num: $readfh/$num\n");
  	if (sysread($readfh,$readdata,$num,length($readdata))==0) {
			&UI::debug("Len: ".length($readdata)."\n");
			return $readdata;
		}
	}
	return $readdata;
}

sub encrypt_message()
{
	# Args: 0: To
	# 1: Message
	local *WFH;
	local *RFH;
	local @arg;
  if (!exists($contacts_sendfrom{$_[0]})) {
    &UI::ynprompt("Request $_[0]'s public key? ", \&request_public_key,
      $_[0]);
    return undef;
  }
	if ($Yeemp::Config::gpg_nolock) {
	  @arg=('--lock-never');
	}
	unlink($clientdir."encrypted_temp.$$");
	local $pid=open2(\*RFH, \*WFH, 'gpg', '--encrypt', 
	  @extra_gpg_options, '--no-default-keyring', '--batch',
		'--keyring', ${clientdir}.'pubring', '--always-trust',
		'--recipient', "$_[0] (from $contacts_sendfrom{$_[0]})", @arg,
		'-o', $clientdir."encrypted_temp.$$");
	print WFH $_[1];
	close(WFH);
	waitpid($pid,0);
	local *TEMPFILE;
	if (!open(TEMPFILE,"<${clientdir}encrypted_temp.$$")) {
	  &UI::stderr("Cannot open ${clientdir}encrypted_temp: $!\n");
	  return '';
	}
	if (!flock(TEMPFILE,&LOCK_EX)) {
	  &UI::stderr("Cannot lock ${clientdir}encrypted_temp.$$\n");
		close(TEMPFILE);
		return '';
	}
	local $/;
	local $ret=<TEMPFILE>;
	if (!unlink($clientdir."encrypted_temp.$$")) {
	  &UI::stderr("Error unlinking ${clientdir}encrypted_temp.$$\n");
	}
	flock(TEMPFILE,&LOCK_UN);
	close(TEMPFILE);
	return $ret;
}

sub encrypt_message_orig()
{
  local *WFH;
  local *RFH;
  local *EFH;
  if (!exists($contacts_sendfrom{$_[0]})) {
    &UI::ynprompt("Generate a keypair for $_[0]? ", \&request_public_key, $_[0]);
    return undef;
  }
  local @arg;
  if ($Yeemp::Config::gpg_nolock) {
    @arg=('--lock-never');
  }
  local $pid=open3(\*WFH,\*RFH,\*EFH,'gpg','--encrypt', @extra_gpg_options,
    '--no-default-keyring', '--keyring', "${clientdir}pubring",
    '--always-trust','--recipient',"$_[0] (from $contacts_sendfrom{$_[0]})",
	@arg);
#  local $message=$_[1];
	local $crypted=&flowinout(RFH,WFH,$_[1]);
#	local $lastfile=select(WFH);
#	$|=1;
#	select(RFH);
#	$|=1;
#	select($lastfile);
#	while (length($message)) {
#	  local $wfd=''; local $rfd=''; local $/;
#	  vec($wfd,fileno(WFH),1)=1;
#		vec($rfd,fileno(RFH),1)=1;
#		select($rfd,$wfd,undef,undef);
#		if (vec($rfd,fileno(RFH),1)) {
#		  sysread(RFH,$crypted,1024,length($crypted));
#		}
#		if (vec($wfd,fileno(WFH),1)) {
#		  local $len;
#		  if (length($message)<4096) {
#			  $len=length($message);
#			} else {
#			  $len=4096;
#			}
#		  local $wrote=syswrite(WFH,$message,$len);
#			if (!defined($wrote)) {
#				&UI::stderr("ERror $! writing to WFH.\n");
#			}
#			substr($message,0,$wrote)='';
#		}
#		&UI::debug("Bleem.\n");
#	}
#	close(WFH);
#	{ local $/; $crypted.=<RFH>; }
#	Old code:
#  select((select(WFH), $|=1)[0]);
#  print WFH $_[1];
#  close(WFH);
#  {
#    local($/);
#    $crypted=<RFH>;
#  }
  if ($crypted eq '') {
    &UI::stderr("I don't have $_[0]'s public key.  Requesting it.\n");
    &request_public_key($_[0]);
  }
  close(WFH);
  close(RFH);
  { local $/;
    &UI::stderr(<EFH>);
  }
  close(EFH);
#  if (defined($oldsigpipe)) {
#    $SIG{PIPE}=$oldsigpipe;
#  } else {
#    delete($SIG{PIPE});
#  }
  return $crypted;
}

sub handle_rules
{
#  print "Server returns rules string $_[0].\n";
}

sub read_channel_init()
{
  local $line;
  while ($line=channel_readln($_[0])) {
    if ($line=~/^9: (.*?)$/) {
      local $rules=$1;
      &handle_rules($rules);
    } elsif ($line=~/^1: (.*?)$/) {
      $channel_expiry{$_[0]}=0;
      return 1;
    } elsif ($line=~/^0: (.*?)$/) {
      channel_close($_[0]);
      return 0;
    } else {
      &UI::stderr("Unhandled line '$line' read.\n");
      channel_close($_[0]);
      return 0;
    }
  }
  return 0;
}

sub open_channel()
{
  if (channel_is_open(CHANNEL)) {
    channel_close(CHANNEL);
  }
  if ($Yeemp::Config::server_channel_local) {
    &UI::stderr("Local server channel unsupported at present.\n");
    return 0;
  } elsif (exists($Yeemp::Config::offline{$Yeemp::Config::server_name})) {
    return 0;
  } else {
    if (Yeemp::Channel::channel_open(CHANNEL, $Yeemp::Config::server_name)) {
      return &read_channel_init(CHANNEL);
    } else {
      $Yeemp::Config::offline{$Yeemp::Config::server_name}=1;
      &UI::stderr("Setting $Yeemp::Config::server_name offline.\n");
    }
  }
  return 0;
}

sub reconnect_to_server()
{
  &Yeemp::Channel::channel_close(CHANNEL);
  if (&open_channel()) {
    &login_to_server();
  }
}

sub lazy_open_channel()
{
  if (channel_is_open(CHANNEL)) {
    local @resp=channel_command(CHANNEL,"Ping");
    if (defined($resp[0]) && ($resp[0] eq '1')) {
      return 1;
    }
    channel_close(CHANNEL);
  }
  return &open_channel();
}

sub send_message($$$)
{
  # Args:
  # 0: Recipient.
  # 1: Message
  # 2: Expiry (Optional)
  local $to=$_[0];
  local $destdomain;
  if ($to=~/^(.*?)~(.*?)$/) {
    $destdomain=$2;
  }
  if (!channel_open(MSGCHAN,$destdomain)) {
    &UI::stderr("Error opening message channel to $destdomain.\n");
    $Yeemp::Config::offline{$destdomain} = 1;
    return 0;
  }
  if (!&read_channel_init(MSGCHAN)) {
    &UI::stderr("Error initializing channel.\n");
    return 0;
  }
  if (defined($_[2])) {
    local @resp=channel_command(MSGCHAN,"Expiry: $_[2]");
    if ($resp[0] ne '1') {
      &UI::stderr("Error setting channel expiry.\n");
      channel_close(MSGCHAN);
      return 0;
    }
  }
  local @resp=channel_command(MSGCHAN, "To: $to");
  if (defined($resp[0]) && ($resp[0] eq '2')) {
    local $munged=$_[1];
    transport_munge($munged);
    channel_write(MSGCHAN,"$munged\012.\012");
    @resp=channel_read_response(MSGCHAN);
    if ($resp[0] eq '1') {
      channel_close(MSGCHAN);
      return 1;
    } else {
      &UI::stderr("Final code: Error spooling for $to: ",@resp,"\n");
    }
  } else {
    &UI::stderr("Error: Cannot send message to $to.\n");
    if ($Yeemp::Config::debug) {
      local $^W=0;
      &UI::stderr("Response was @resp\n");
    }
  }
  channel_close(MSGCHAN);
  return 0;
}

sub spool_munge()
{
  local $block=$_[0];
  if (defined($block)) {
    $block=~s/^/ /gm;
  }
  return $block;
}

sub spool_message()
{
  local *FILE;
  if (open (FILE,">>$_[0]")) {
    if (flock(FILE,LOCK_EX)) {
      seek(FILE,0,2);
      if ((print FILE "To: $_[1]\n") && (print FILE &spool_munge($_[2])."\n")) {
#        &UI::debug("Message spooled in postponed file.\n");
      } else {
        &UI::stderr("Unable to write message to postponed spool.\n");
      }
      flock(FILE,LOCK_UN);
    } else {
      &UI::stderr("Unable to flock postponed spool $_[0].\n");
    }
    close(FILE);
  } else {
    &UI::stderr("WARNING:  Unable to open message spool $_[0]\n");
  }
}

sub spool_postponed_message_crypted()
{
  &spool_message($Yeemp::Config::postponed_spool_crypted,$_[0],$_[1]);
}

sub deliver_message($$;$)
{
  # Args: 0: Recipient, 1: Message, 2: (optional, default 0): Expiry.
  local $mess=&encrypt_message($_[0],$_[1]);
  if (defined($mess) && ($mess ne '')) {
#		This was causing the corruption.
#    transport_munge($mess);
    &lazy_open_channel();
    if (&send_message($_[0],$mess,$_[2])) {
      return 1;
    }
    if ((!defined($_[2])) || ($_[2]==0)) {
      &UI::stderr("Unable to send message; spooling for later.\n");
      &spool_postponed_message_crypted($_[0],$mess,$_[2]);
    } else {
      &UI::debug("Not spooling transient message.\n");
    }
  } else {
    &UI::stderr("Unable to encrypt message for $_[0].\n");
    &spool_postponed_message_bare($_[0],$_[1],$_[2]);
  }
  return 0;
}

sub initialize_client()
{
  local %args=@_;
  # Accepted args: clientdir - set $clientdir explicitly.
  # If not set, will be [homedir]/.yeemp/ .
  &Yeemp::Common::initialize_common(@_);
  &clean_environment();
  # Make waitpid() unnecessary.
  $SIG{CHLD}='IGNORE';
  if (defined(%args)) {
    if (exists($args{clientdir})) {
      $clientdir=$args{clientdir};
      if ($clientdir!~/\/$/) {
        $clientdir.='/';
      }
#      eval("use lib '$clientdir';");
    }
  }
  if (!defined($clientdir)) {
    @pwent=getpwuid($<);
    $user=$pwent[0];
    $homedir=$pwent[7];
    $clientdir="$homedir/.yeemp/";
  }
  $Yeemp::Config::postponed_spool_bare=$clientdir.'postponed_bare';
  $Yeemp::Config::postponed_spool_crypted=$clientdir.'postponed_crypted';
  $Yeemp::Config::certcache_dir=$clientdir.'certs';
#  if (defined($Yeemp::Config::default_machine) && $user) {
#    $Yeemp::Config::my_address="$user~$Yeemp::Config::default_machine";
#  }
  umask(0077); #Privacy is your friend.
  $ENV{GNUPGHOME}=$clientdir;

  if ( ! (-d $clientdir)) {
    &UI::debug("$clientdir doesn't exist.  Creating it with mode 0700.\n");
    mkdir($clientdir,0700);
  }
  if ( ! (-d $Yeemp::Config::certcache_dir)) {
    &UI::debug("Creating $Yeemp::Config::certcache_dir.\n");
    mkdir($Yeemp::Config::certcache_dir,0700);
  }
  load_client_config();
  local *RING;
  local @stat=stat($clientdir.'pubring');
  if (!defined(@stat)) {
    &UI::stderr("No public key ring.  Creating it.\n");
    open (RING,">>${clientdir}pubring");
    close(RING);
    chmod (0600,"${clientdir}pubring");
  } elsif ($stat[2] & 0077) {
    if (!$Yeemp::Config::allowbadperms) {
      &UI::stderr("Your public key ring is accessible to others.\n");
      &UI::stderr("If other users read it, this will allow them to send messages\nthat appear to come from you to the contacts in the list.\n");
      &UI::stderr("Fixing the permissions.\n");
      chmod(0600,"${clientdir}pubring");
    }
  }

  @stat=stat("${clientdir}secring");
  if (!defined(@stat)) {
    umask(077);
    &UI::stderr("No secret key ring.  Creating it.\n");
    open (RING, ">>${clientdir}secring");
    close(RING);
    chmod (0600,"${clientdir}secring");
  } elsif ($stat[2] & 0077) {
    if (!$Yeemp::Config::allowbadperms) {
      &UI::stderr("ACK!  Your secret key ring is accessible to others.\nThis is  very, very bad.  If a malicious individual or program reads\nyour secret key ring, they can decrypt all messages addressed to you.\n");
      &UI::stderr("Aborting Yeemp.\n");
      exit;
    }
  }

  # Stick any options you want applied to *ALL* invocations of gpg
  # in this array.
  #  Ignoring time conflict because exchanging messages with machines
  # that have brain-damaged ideas of 'now' is likely more important
  # than security breaches driven by time travel.
  @extra_gpg_options=('--ignore-time-conflict');

  mkdir($clientdir.'cache',0700);
  mkdir($clientdir.'logs',0700);
}

sub cache()
{
  # Args: 0: Sender address.  1: Filename.  2: File content.
  local *FILE;
  local $username;
  local $file=$_[1];
  if ($_[0]=~/^([-a-zA-Z~0-9._]+)$/) {
    $username=$1;
  } else {
    &UI::stderr("Username is not a safe directory name; message not cached.\n");
    return;
  }
  if ($_[1]=~/\.\.\//) {
    &UI::stderr("Untrustworthy filename; message not cached.\n");
    return;
  }
  my @dirs = split(/\//,$_[1]);
  my $dir="$clientdir/cache/$username";
#  if ($_[1]=~/\/([-a-zA-Z_.]+)$/) {
#    $file=$1;
#  } else {
#  }
  mkdir("$clientdir/cache/$_[0]",0700);
  while ($#dirs) {
    $dir.="/".shift(@dirs);
    mkdir($dir,0700);
  }
  open(FILE,">$clientdir/cache/$username/$file") or
    UI::stderr("Error opening $clientdir/cache/$username/$file for writing.\n");
  print FILE $_[2];
  close(FILE);
}

sub log_message()
{
  # Args:
  # 0: From.
  # 1: Message.
  local *FILE;
  local $retval=0;
  local $logfile;
  local $from=$_[0];
  if (!defined($from)) {
    $from='Undefined';
  }
  if ($from=~/^([.-_0-9a-zA-Z~]+)$/) {
    $logfile=$1;
    if (open(FILE,">>$clientdir/logs/$logfile")) {
      if (flock(FILE,LOCK_EX)) {
        $retval=(print FILE $_[1]);
        flock(FILE,LOCK_UN);
      }
      close(FILE);
    }
  } else {
    &UI::stderr("Invalid characters in message from '$from'.  Unable to log.\n");
  }
  return $retval;
}

sub init_global_config()
{
  package Yeemp::Config;
  $cache_status=1;            # Status caching is on by default.
  $status_change_expiry=150;  # Status changes expire in two and a half minutes
  $default_presence_mode='confirm';  #  Confirm before announcing status.
}

sub enable_esd_playback()
{
  UI::stdout("ESD will be used if it's running.\n");
  $Yeemp::Config::use_esd = 1;
}

sub enable_ogg_playback()
{
  UI::stdout("Ogg playback enabled.\n");
  UI::ynprompt(
  "ESD, the Enlightened Sound Daemon, permits multiple applications to
use the sound hardware simultaneously.  However, monitoring ESD output
is trivial for anyone on your machine.  (Or between you and the sound
server, if using esd's net-play features.)

Use ESD if it's running? ", \&enable_esd_playback, undef);
  $Yeemp::Config::play_oggs = 1;
}

sub determine_media_settings()
{
  UI::ynprompt("Currently, Yeemp can use ogg123 to play sound files supplied by your
contacts.  This is currently used for status-change sounds, and will be
used for audio messaging.  This presents a potential security risk for
two reasons: first, contact-supplied data will be passed to ogg123 without
any sanity checks; if there are exploitable security flaws in ogg123 or
your sound driver, this will expose them.  Also, audio or electromagnetic
surveillance will be able to intercept any audio played.

Play Ogg Vorbis audio files?",
    \&enable_ogg_playback, undef);
}

sub load_client_config()
{
  &init_global_config();
  package Yeemp::Config;
  # We want the config in the Yeemp::Config package
  # Grab global config from @INC
  eval("require 'yeemp.cfg';");
  if (!eval ( 'require "${Yeemp::Client::clientdir}client.cfg";' )) {
    package Yeemp::Client;
    local *CFG;
    if (sysopen(CFG,"${clientdir}client.cfg",O_WRONLY|O_CREAT|O_EXCL)) {
      if (flock(CFG,LOCK_EX)) {
        print CFG '$cache_status=1;
$status_change_expiry=150;
$keyauth_mode="anon";
$client_mode="remote";
$default_presence_mode="confirm";
$default_status="Connected";
$allowbadperms=0;
$announce_status=1;
$maxopenwindows=10;
$maxportraitsize=256;
# Warning: Audio can be intercepted by a decent microphone or EM monitor.
$play_oggs=0;
$beep_on_statchange=0;
$request_media_files=1;
# By default, clients will only request appropriate media.  This can be
# used to determine whether a user is using the X or console clients, as
# the console client will not request graphics automatically.
# if force_media_request is set to 0, the X client will not request graphics.
# if force_media_request is set to 1, the console client will request graphics.
#$force_media_request=0;
#  Warning: esd output can be sniffed trivially
$use_esd=0;
1;
';
        flock(CFG,LOCK_UN);
      } else {
				&UI::stderr("Error: Couldn't lock ${clientir}client.cfg\n");
			}
      close(CFG);
    } else {
			&UI::stderr("Error writing ${clientdir}client.cfg\n");
		}
    package Yeemp::Config;
    require "${Yeemp::Client::clientdir}client.cfg";
    package Yeemp::Client;
    UI::init_preferences();
    Yeemp::Client::determine_media_settings();
  } else {
    if ((!defined($server_name)) || (!defined($my_address))) {
      UI::init_preferences();
      Yeemp::Client::determine_media_settings();
    }
		if ((!defined($maxopenwindows)) || ($maxopenwindows<1)) {
		  &UI::stdout("Max open windows set to 10.\n");
		  $maxopenwindows=10;
		}
  }
  $Yeemp::Client::my_status=$Yeemp::Config::default_status;
  return 1;
}

sub clean_environment()
{
  $ENV{PATH}='/bin:/usr/bin:/usr/local/bin';
  delete($ENV{ENV});
}

sub cb_contactlist_add_deekoo($)
{
  add_contact('deekoo~tentacle.net', 'Deekoo');
}

sub load_contact_list()
{
  local *CONTACTS;
  if (!open(CONTACTS,"<${clientdir}contacts")) {
    local *DIR;
    if (opendir(DIR,$clientdir)) {
      closedir(DIR);
    } else {
      if (!mkdir($clientdir,0700)) {
        $load_contact_list_error = "mkdir $clientdir failed.";
        return 0;
      }
    }
    if (!open (CONTACTS,">>${clientdir}contacts")) {
      $load_contact_list_error = "Cannot create contact list.";
      return 0;
    }
    close(CONTACTS);
    UI::ynprompt(
      "Creating new contact list.  Put deekoo~tentacle.net on it?",
      \&cb_contactlist_add_deekoo, undef);
    if (!open (CONTACTS,"<${clientdir}contacts")) {
      $load_contact_list_error="Cannot open contact list";
      return 0;
    }
  }
  local $line;
  local @line;
  local $contact;
  while ($line=<CONTACTS>) {
    chomp($line);
    if (defined($line) && ($line ne '')) {
      @line=split('=',$line);
      if ($line[0] eq 'Contact') {
        $contact=$line[1];
        $contacts{$contact}=1;
        $contacts_status_updated{$contact}=0;
        &load_cached_status($contact);
      } elsif ($line[0] eq 'Presence') {
        $contacts_presence{$contact}=$line[1];
#      } elsif ($line[0] eq 'Require_Signature') {
#        $contacts_require_signatures{$contact}=$line[1];
      } elsif ($line[0] eq 'Recvfrom') {
        $contacts_recvfrom{$contact}=$line[1];
      } elsif ($line[0] eq 'Sendfrom') {
        $contacts_sendfrom{$contact}=$line[1];
      } elsif ($line[0] eq 'Shortname') {
        $contacts_shortnames{$contact}=$line[1];
      } else {
        &UI::stderr("WARNING: Unhandled contact line '$line'\n");
      }
    }
  }
  close(CONTACTS);
  return 1;
}

sub shortname_to_address($)
{
  local $k;
  foreach $k (keys(%contacts_shortnames)) {
    if ($contacts_shortnames{$k} eq $_[0]) {
      return $k;
    }
  }
  return undef;
}

sub load_cached_status()
{
  # $_[0]: Username to load status for.
  local *FILE;
  $contacts_status{$contact}='[unknown]';
  if (open(FILE,"<${clientdir}cache/$_[0]/status")) {
    local $/;
    local $munged=<FILE>;
    if ($munged=~/^(.*?)\012 (.*)$/s) {
      local $unmunged=$2;
      $unmunged=~s/^ //gm;
      if ($unmunged=~/^Status: (.*?)$/m) {
        $contacts_status{$contact}=$1;
      }
    }
    close(FILE);
  }
}

sub set_cfg_line($$$)
{
  local $var=$_[1];
  local $val=$_[2];
	if (!defined($val)) {
		$val="''";
	}
  if ($_[0]=~/^\$$var=(.*?);$/m) {
    $_[0]=~s/^\$$var=(.*?);$/\$$var=$val;/m;
  } else {
    $_[0]=~s/^1;$/\$$var=$val;\n1;/m;
  }
}

sub save_client_cfg()
{
  local *CFG;
  open(CFG,"+<${clientdir}client.cfg") || return 0;
  local $/;
  local $cfg=<CFG>;
  local $oldcfg=$cfg;
  &set_cfg_line($cfg, 'cache_status', $Yeemp::Config::cache_status);
  &set_cfg_line($cfg, 'status_change_expiry', $Yeemp::Config::status_change_expiry);
  &set_cfg_line($cfg, 'server_name', "\"$Yeemp::Config::server_name\"");
  &set_cfg_line($cfg, 'my_address', "\"$Yeemp::Config::my_address\"");
  &set_cfg_line($cfg, 'default_presence_mode', "\"$Yeemp::Config::default_presence_mode\"");
  if (!defined($Yeemp::Config::announce_status)) {
    $Yeemp::Config::announce_status=0;
  }
  &set_cfg_line($cfg, 'announce_status', $Yeemp::Config::announce_status);
	&set_cfg_line($cfg, 'maxopenwindows', $Yeemp::Config::maxopenwindows);
	&set_cfg_line($cfg, 'maxportraitsize', $Yeemp::Config::maxportraitsize);
  set_cfg_line($cfg, 'play_oggs', $Yeemp::Config::play_oggs);
  set_cfg_line($cfg, 'use_esd', $Yeemp::Config::use_esd);
  if ($cfg ne $oldcfg) {
    seek(CFG,0,SEEK_SET);
    print CFG $cfg;
    truncate(CFG,tell(CFG));
  }
  close(CFG);
}

sub save_contact_list()
{
  local *CONTACTS;
  local $k;
  local $ret=0;
  $file='';
  foreach $k (keys(%contacts)) {
    $file.="Contact=$k\n";
    if (defined($contacts_shortnames{$k})) {
      $file .= "Shortname=$contacts_shortnames{$k}\n";
    }
#    if (defined($contacts_require_signatures{$k})) {
#      $file.="Require_Signature=$contacts_require_signatures{$k}\n";
#    }
    if (defined($contacts_recvfrom{$k})) {
      $file.="Recvfrom=$contacts_recvfrom{$k}\n";
    }
    if (defined($contacts_sendfrom{$k})) {
      $file.="Sendfrom=$contacts_sendfrom{$k}\n";
    }
    if (exists($contacts_presence{$k})) {
      $file.="Presence=$contacts_presence{$k}\n";
    }
  }
  if (open(CONTACTS,">${clientdir}contacts.new")) {
    if (flock(CONTACTS,LOCK_EX)) {
      if (print CONTACTS $file) {
        rename("${clientdir}contacts.new", "${clientdir}contacts");
        $ret=1;
      }
      flock(CONTACTS,LOCK_UN);
    } else {
      print "Error: Cannot lock contact list.\n";
    }
    close(CONTACTS);
  }
  return $ret;
}

sub set_contact_sendfrom()
{
  if (!exists($contacts{$_[0]})) {
    $contacts{$_[0]}=undef;
  }
  if (!exists($contacts_sendfrom{$_[0]})) {
    $contacts_sendfrom{$_[0]}=$_[1];
    save_contact_list();
    return 1;
  } else {
    if ($contacts_sendfrom{$_[0]} eq $_[1]) {
      return 1;
    }
    print "WARNING: contacts_sendfrom{$_[0]} already set.  Not setting ($_[1] discarded.)\n";
  }
  return 0;
}

sub send_arbitrary_file()
{
	# 0: Hashref
	local %args=%{$_[0]};
	# Args: sendas: Name to send file as
	if ($args{sendas} eq '') {
	  &UI::stderr("BUG: Unable to send file: name to send as not specified.\n");
		return;
	}
	if ($args{to} eq '') {
	  &UI::stderr("BUG: Recipient of file not specified.\n");
		return;
	}
	if ($args{file} eq '') {
	  &UI::stderr("BUG: File to send not specified.\n");
		return;
	}
	local *FILE;
	if (!open(FILE,"<$args{file}")) {
	  &UI::stderr("Can't read $args{file}.\n");
		return;
	}
	local $oldslash=$/;
	undef($/);
	local $message=
	  "From: $Yeemp::Config::my_address\012File: $args{sendas}\012\012".<FILE>;
	$/=$oldslash;
	close(FILE);
	&UI::debug("About to deliver.\n");
	&deliver_message($args{to},$message);
}

sub send_file()
{
  local $who=$_[0]; # Arg 0: Recipient.
  local $file=$_[1]; # Arg 1: Filespec.
  local $message;
  # Check for magic filespecs.
  if ($file eq '/public_key') {
    local $pubkey=&extract_receiving_public_key($who);
    if (defined($pubkey)) {
      $message = "From: $Yeemp::Config::my_address\012File: /public_key\012\012$pubkey\012";
    }
  } elsif ($file eq '/status') {
    $message = "From: $Yeemp::Config::my_address\012Status: $my_status\012\012";
  } elsif (-f $file) {
	  &UI::read_line("Send file as: ", 'send', sub { 
			local %args=('sendas' => $_[0],
				'to' => ${$_[1]}[0],
				'file' => ${$_[1]}[1]);
			&send_arbitrary_file(\%args);
		}, $file, \@_);
  } else {
		&UI::stderr("$file not found.\n");
	}
  if (defined($message)) {
    &deliver_message($who,$message);
  } else {
    &UI::stderr("DEBUG: Bug of some sort, no file sent.\n");
  }
}

sub channel_read_block_unchanged()
{
  local ($line='',$block='');
  while (($line ne ".") && defined($line)) {
    $line=channel_readln(CHANNEL);
    if ($line!~/^\./) {
      $block.=$line;
      $block.="\012";
    } else {
      if ($line ne ".") {
        $block.=substr($line,1);
        $block.="\012";
      }
    }
  }
  $block=~s/\012$//s;
#  print &clean("About to return '$block'\n");
  return $block;
}

sub sign_challenge()
{
  # Args: 0: Server name, 1: Challenge.
  local *WFH;
  local *RFH;
  local *EFH;
  open3(\*WFH,\*RFH,\*EFH,'gpg','--detach-sign', '--no-default-keyring',
    '--secret-keyring', "$clientdir/login-key-$Yeemp::Config::my_address",
    '--keyring', "$clientdir/login-key-$Yeemp::Config::my_address.pub",
    @extra_gpg_options);
  print WFH "$_[1]";
  close(WFH);
  local $/;
  local $fds=0;
  vec($fds,fileno(RFH),1)=1;
  vec($fds,fileno(EFH),1)=1;
  local $sig='';
  local $err='';
  while (defined(fileno(RFH)) && defined(fileno(EFH))) {
    local $readfds;
    local $nfound=select($readfds=$fds,undef,undef,undef);
    if (vec($readfds,fileno(RFH),1)) {
      local $frag=<RFH>;
      if (defined($frag)) {
        $sig.=$frag;
      } else {
        vec($fds,fileno(RFH),1)=0;
        close(RFH);
      }
    }
    if (vec($readfds,fileno(EFH),1)) {
      local $frag=<EFH>;
      if (defined($frag)) {
        $err.=$frag;
      } else {
        vec($fds,fileno(EFH),1)=0;
        close(EFH);
      }
    }
  }
  if ($err ne '') {
    &UI::stderr("Debug: sign_challenge: GPG errors:\n$err.\n");
  }
  if ($sig ne '') {
    return $sig;
  }
  return undef;
}

sub login_to_server()
{
  local $challenge;
  local @resp=channel_command(CHANNEL,"Login: $Yeemp::Config::my_address");
#  &UI::stderr("Debug: @resp\n");
  if (defined($resp[0]) && ($resp[0] eq '3')) {
    $challenge=&channel_read_block_unchanged();
    local $sig=&sign_challenge($Yeemp::Config::my_address, $challenge);
    if (defined($sig)) {
      @resp=channel_command(CHANNEL,"Signature");
      if ($resp[0] eq '2') {
        transport_munge($sig);
        channel_write(CHANNEL,"$sig\012.\012");
        local $line=channel_readln(CHANNEL);
        if ($line =~/^1: /) {
          $channel_logged_in=1;
          local @resp=channel_command(CHANNEL,"Notify");
          if ($resp[0] ne 1) {
            &UI::stderr("Notify request failed: @resp\n");
					}
          return 1;
        } elsif ($line=~/^0: /) {
          chomp($line);
          &UI::stderr("Login failed: Bad signature.\n");
          &UI::stderr("(Details: $line)\n");
        } else {
          &UI::stderr("Invalid response to signature command: $line\n");
        }
      } elsif ($line=~/^0: /) {
        &UI::stderr("Signature command denied.\n");
      } else {
        &UI::stderr("Invalid response to signature command: $line\n");
      }
    }
  } elsif ($resp[0] eq '0') {
    &UI::stderr("Server rejected our login attempt.\n");
    &UI::stderr("(Details: ",@resp,")\n");
  } else {
    &UI::stderr("Bad response to login command: ",@resp,"\n");
  }
}

sub lazy_login_to_server()
{
  if ($channel_logged_in) {
    return 1;
  }
  return &login_to_server();
}

sub get_new_message()
{
  local @resp=channel_command(CHANNEL,"Retrieve");
  if (!defined($resp[0])) {
    &UI::stderr("Invalid response ($resp[1])\n");
  } elsif ($resp[0] eq '3') {
    local $mess=&channel_read_block_unchanged();
    return $mess;
#  } else {
#    &UI::stderr("Unable to get a message: server responds @resp.\n");
  }
  return undef;
}

BEGIN
{
#  &initialize_client();
  return 1;
}

sub resend_postponed_bare ()
{
  local %headers=&split_headers($_[0]);
  local $key=&extract_sending_public_key($headers{To});
  if (defined($key) && ($key ne '')) {
    local $crypted=&encrypt_message($headers{To},$_[1]);
    if ($crypted) {
      if (&lazy_open_channel()) {
        &send_message($headers{To},$crypted) ||
          &spool_postponed_message_crypted($headers{To},$crypted);
      } else {
        &UI::stdout("Unable to open server channel.\n");
        &spool_postponed_message_crypted($headers{To},$crypted);
      }
      return 1;
    }
  } else {
    &bunch_output("Postponed message for $headers{To} is remaining postponed until they\nsend you their public key.\n");
#    &UI::stdout("Postponed message for $headers{To} is remaining postponed until they\nsend you their public key.\n");
  }
  return 0;
}

sub bunch_output()
{
  $bunched_output{$_[0]}='';
}

sub resend_bare_postponed_spool()
{
  local $start=0;
  local $end=0;
  local *SPOOL;
  if (open(SPOOL, "+<$Yeemp::Config::postponed_spool_bare")) {
    if (flock(SPOOL,LOCK_EX|LOCK_NB)) {
      local $data='';
      local $header='';
	  &UI::stdout("Bare postponed messages found.\n");
      while ($line=<SPOOL>) {
        if (substr($line,0,1) eq ' ') {
          $data.=$line;
        } elsif ($line ne '') {
          if ($data ne '') {
            local $unmunged=&unmunge_stored($data);
            if (&resend_postponed_bare($header,$unmunged)) {
#              print "Snipping postponed.\n";
              &file_snip(*SPOOL{IO},$start,$end);
              $end=$start;
            }
            $start=$end;
            $data='';
            $header='';
          }
          $header.=$line;
        }
        $end=tell(SPOOL);
      }
      seek(SPOOL,0,SEEK_END);
      if ((tell(SPOOL)==0) && ((stat($Yeemp::Config::postponed_spool_bare))[7]==0)) {
        unlink($Yeemp::Config::postponed_spool_bare);
      }
      flock(SPOOL,LOCK_UN);
    } else {
      &UI::stderr("Error locking bare message spool.\n");
    }
    close(SPOOL);
  }
  &UI::stdout(keys(%bunched_output));
  undef(%bunched_output);
}

sub resend_crypted_postponed_spool()
{
  local $start=0;
  local $end=0;
  local *SPOOL;
  if (open(SPOOL, "+<$Yeemp::Config::postponed_spool_crypted")) {
    if (flock(SPOOL,LOCK_EX|LOCK_NB)) {
      if (&lazy_open_channel()) {
        if (&lazy_login_to_server()) {
          local $data='';
          local $header='';
          &UI::stdout("Encrypted postponed messages found.\n");
          local $notdone=1;
          while ($notdone) {
            $line=<SPOOL>;
            if (defined($line) && (substr($line,0,1) eq ' ')) {
              $data.=$line;
            } else {
              if ($data ne '') {
                local $unmunged=&unmunge_stored($data);
                if ($header=~/^To: (.*?)$/m) {
                  if (&send_message($1,$unmunged)) {
#                    print "Snipping postponed ($start, $end).\n";
                    &file_snip(*SPOOL{IO},$start,$end);
                    seek(SPOOL,0,SEEK_SET);
                    $start=0;
                    $end=0;
                  }
                } else {
                  &UI::stderr("Warning: No to header ($header)\n");
                  &clean_print("Block: $unmunged.\n");
                }
#                $start=$end;
                $data='';
                $header='';
              }
              if (defined($line)) {
                $header.=$line;
              } else {
                $notdone=0;
              }
            }
            $end=tell(SPOOL);
          }
          seek(SPOOL,0,SEEK_END);
           if ((tell(SPOOL)==0) && ((stat($Yeemp::Config::postponed_spool_crypted))[7]==0)) {
#            print "Unlinking spool.\n";
            unlink($Yeemp::Config::postponed_spool_crypted);
          }
        } else {
          &UI::stderr("Unable to login to server.\n");
        }
      } else {
        &UI::stderr("Unable to open channel.\n");
      }
      flock(SPOOL,LOCK_UN);
    } else {
      &UI::stderr("Error locking crypted message spool.\n");
    }
    close(SPOOL);
  }
}

sub edit_contact()
{
  if (exists($contacts{$_[0]})) {
    &UI::edit_contact_by_address($_[0]);
    return;
  } else {
    local $cont;
    foreach $cont (keys(%contacts_shortnames)) {
      if (lc($contacts_shortnames{$cont}) eq lc($_[0])) {
        &UI::edit_contact_by_address($cont);
        return;
      }
    }
  }
  &UI::stderr("Cannot find a contact called '$_[0]'.\n");
}

# What does this do?
@ISA=qw(Exporter);

#	$clientdir
#	$my_address
#	$config_cache_status
#	$config_client_mode
#	$config_postponed_spool_bare
#	$config_postponed_spool_crypted
#	$config_server_name
#	$config_status_change_expiry

@EXPORT=qw(
	$sigpipe_raised
	%contacts
	%contacts_recvfrom
	%contacts_sendfrom
	%contacts_status
	%contacts_status_updated
	cache
	check_login_key
	check_postponed_spools
	create_recipient_public_key
	deliver_message
	encrypt_message
	extract_sending_public_key
	import_new_messages
	load_contact_list
	log_message
	login_to_server
	open_channel
	save_contact_list
	send_message
	set_contact_sendfrom
	unmunge_stored
	untaint_environment
	update_status
	);
1;
