#
#  Copyright 2001 Nathaniel E. Bray/Deekoo L.
#
#  See LICENSE for details.
#
package Yeemp::SCP;

use Expect;

my $unauthed_regex = 
    '(?s)The authenticity of host \'(.*?) \((.*?)\)\' can\'t be established\.\015\012(.*?) key fingerprint is (.*?).\015\012Are you sure you want to continue connecting? ';
my $password_regex =
  '(?s)(.*?)\@(.*?)\'s password: ';

sub new(%)
{
  shift @_;
  my %self = @_;
  ($self{destuser},$self{desthost}) = split('@', $self{destacct}, 2);
  $self{'exp'} = Expect->spawn('scp', ($self{sourcefile},
    $self{destacct}.':'.$self{destfile}));
  $self{'exp'}->log_user(0);
  $self{'state'} = 0;
  return bless (\%self, Yeemp::SCP);
}

sub abort($)
{
  &UI::stderr("SCP aborted: Unexpected response from slave client.");
  $_[0]->dispose;
}

sub cycle()
{
  local @got = ${$_[0]}{"exp"}->expect(30,
    [$unauthed_regex, sub {}],
    [$password_regex, sub {}]
    );
  if (!defined($got[0])) {
    if ($got[1]=~/^2:EOF/) {
      $_[0]->dispose;
    } else {
      &UI::stderr("Unexpected error in Yeemp::SCP->cycle: @got\n");
    }
  } elsif ($got[0] == 1) {
    if (${$_[0]}{state} != 0) {
      scp_abort($_[0]);
      return;
    }
    $got[2]=~/$unauthed_regex/;
    local $host=$1;
    local $ip=$2;
    local $keytype=$3;
    local $fingerprint=$4;
    if ($host ne ${$_[0]}{desthost}) {
      scp_abort($_[0]);
      return;
    }
    &UI::ynprompt("The authenticity of $host ($ip) can't be established.
    $keytype key fingerprint is $fingerprint.
    Accept this host key as valid for $host ($ip)?",
    sub { my $self=shift;
          my $exp=${$self}{'exp'};
          print $exp "yes\n";
          $self->cycle;
          }, $_[0],
    sub {
      my $self=shift;
      my $exp = ${$self}{'exp'};
      print $exp "no\n";
      $self->dispose; });
    ${$_[0]}{state} = 1;
  } elsif ($got[0] == 2) {
    if (${$_[0]}{state}==0 or ${$_[0]}{state}==1) {
      &UI::passwordprompt("Password for ${$_[0]}{destacct}",
        sub {
        $self = $_[1];
        my $exp = ${$self}{'exp'};
        print $exp "$_[0]\n";
        $self->cycle; },
        $_[0]);
    }
  }
}

sub dispose($)
{
  my $self=shift;
  my $pid = ${$self}{'exp'}->exp_pid;
  &{${$self}{callback}}();
  if ($pid) {
    eval {
      local $SIG{ALRM}=sub {die};
      alarm 10;
      if (waitpid($pid, 0)!=-1) {
        if (eval {
          alarm 10;
          kill('TERM', $pid);
          return waitpid($pid, 0);
        }!=-1) {
          kill('KILL', $pid);
        }
      }
      alarm 0;
    }
  }
  undef(${$self}{'exp'});
  undef(%$self);
}

1;
