package UI;
use Term::ReadKey;
#use utf8;

%VT100 = (
  'cpr'   => "\x1b[6n",
  'rst'   => "\x1b[0m",
  'invis' => "\x1b[0;30;40m",
  'rep0'  => "\x1b[",
  'repy'  => '([0-9]+)',
  'rep1'  => ';',
  'repx'  => '([0-9]+)',
  'rep2'  => 'R',
  'tocol0'=> "\015"
  );

@funckeys = (
  'Escape', 'Pause',
  'Insert', 'Delete', 'Home', 'End', 'Page Up', 'Page Down',
  'Up', 'Down', 'Left', 'Right', 'Center',
  'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12',
  'F13', 'F14', 'F15', 'F16', 'F17', 'F18', 'F19', 'F20',
  'Num Lock', 'Num /', 'Num *', 'Num -', 'Num +', 'Num Enter',
  'Num 0', 'Num 1', 'Num 2', 'Num 3', 'Num 4', 'Num 5', 'Num 6',
  'Num 7', 'Num 8', 'Num 9', 'Num .'
  );

%funckeys_default = (
  # Function key strings.  undef marks a prefix.
  "\x1b"      =>  'Escape',
  "\x1b[1~"   =>  'Home',
  "\x1b[2~"   =>  'Insert',
  "\x1b[3~"   =>  'Delete',
  "\x1b[4~"   =>  'End',
  "\x1b[5~"   =>  'Page Up',
  "\x1b[6~"   =>  'Page Down',
  "\x1b[15~"  =>  'F5', # xterm
  "\x1b[17~"  =>  'F6',
  "\x1b[18~"  =>  'F7',
  "\x1b[19~"  =>  'F8',
  "\x1b[20~"  =>  'F9',
  "\x1b[21~"  =>  'F10',
  "\x1b[23~"  =>  'F11',
  "\x1b[24~"  =>  'F12',
  "\x1b[25~"  =>  'F13',
  "\x1b[26~"  =>  'F14',
  "\x1b[27~"  =>  'F15',
  "\x1b[28~"  =>  'F16',
  "\x1b[31~"  =>  'F17',
  "\x1b[32~"  =>  'F18',
  "\x1b[33~"  =>  'F19',
  "\x1b[34~"  =>  'F20',
  "\x1b[A"    =>  'Up',
  "\x1b[B"    =>  'Down',
  "\x1b[C"    =>  'Right',
  "\x1b[D"    =>  'Left',
  "\x1b[E"    =>  'Center',     # xterm
  "\x1b[F"    =>  'End',        # xterm
  "\x1b[G"    =>  'Center',
  "\x1b[H"    =>  'Home',       # xterm
  "\x1b[[A"   =>  'F1',
  "\x1b[[B"   =>  'F2',
  "\x1b[[C"   =>  'F3',
  "\x1b[[D"   =>  'F4',
  "\x1b[[E"   =>  'F5',
  "\x1bOM"    =>  'Num Enter',
  "\x1bOP"    =>  'F1',         # xterm
  "\x1bOQ"    =>  'F2',         # xterm
  "\x1bOR"    =>  'F3',         # xterm
  "\x1bOS"    =>  'F4',         # xterm
  "\x1bOl"    =>  'Num +',
  "\x1bOn"    =>  'Num .',
  "\x1bOp"    =>  'Num 0',
  "\x1bOq"    =>  'Num 1',
  "\x1bOr"    =>  'Num 2',
  "\x1bOs"    =>  'Num 3',
  "\x1bOt"    =>  'Num 4',
  "\x1bOu"    =>  'Num 5',
  "\x1bOv"    =>  'Num 6',
  "\x1bOw"    =>  'Num 7',
  "\x1bOx"    =>  'Num 8',
  "\x1bOy"    =>  'Num 9',
  );

%funckeys_linux = (
  # Console codes which clobber xterm's here.
  "\x1bOP"    =>  'Num Lock',
  "\x1bOQ"    =>  'Num /',
  "\x1bOS"    =>  'Num -',
  "\x1bOR"    =>  'Num *',
);

%funckeys_xterm = (
  # Empty for now
  );

%keytrans = (
  # ASCII values for function keys.
  # Escape is needed if you want to be able to trap a literal escape.
  'Escape'    =>  "\x1b",
  'Delete'    =>  "\x07",
  'Home'      =>  "\x01",
  'End'       =>  "\x05",
  'Left'      =>  "\x02",
  'Right'     =>  "\x06",
  'Num Enter' =>  "\x0a",
  'Num /'     =>  '/',
  'Num *'     =>  '*',
  'Num -'     =>  '-',
  'Num +'     =>  '+',
  'Num 0'     =>  '0',
  'Num 1'     =>  '1',
  'Num 2'     =>  '2',
  'Num 3'     =>  '3',
  'Num 4'     =>  '4',
  'Num 5'     =>  '5',
  'Num 6'     =>  '6',
  'Num 7'     =>  '7',
  'Num 8'     =>  '8',
  'Num 9'     =>  '9',
  'Num .'     =>  '.',
  );

$function_key_timeout = 0.1;  # Timeout.

sub add_funckey_nodes()
{
  my @keys = keys(%funckeys);
  my $k;
  for $k (@keys) {
    my $x=1;
    while ($x<length($k)) {
      if (!exists($funckeys{substr($k,0,$x)})) {
        $funckeys{substr($k,0,$x)} = undef;
      }
      $x++;
    }
  }
}

sub init()
{
  if ($ENV{TERM} eq 'linux') {
    %funckeys = (%funckeys_default, %funckeys_linux);
  } elsif ($ENV{TERM} eq 'xterm') {
    %funckeys = (%funckeys_default, %funckeys_xterm);
  } else {
    print STDERR "Debug: Unsupported terminal $ENV{TERM}.  Falling back to default.\n";
    %funckeys = %funckeys_default;
  }
  add_funckey_nodes();
  ReadMode 4;
  $| = 1;
  init_utf8_settings();
}


sub probe_utf8_settings()
{
  print $VT100{invis};
  print $VT100{tocol0};
  my @pos0 = getpos();
  $singlebyte = 0;
  if ($#pos0==-1) {
    print $VT100{rst}, "Unable to get cursor position.\nAssuming that your term thinks all non-ASCII characters are doublewidth.\n";
    $doublewidth = 0x80;
  } else {
    print "怀";
    @pos1 = getpos();
    print $VT100{tocol0},$VT100{rst};
    my $delt = $pos1[0] - $pos0[0];
    if ($delt == 1) {
      # UTF-8 mode, no doublewidth chars
      $doublewidth = 0;
      print "VT is in UTF-8 mode, without doublewidth characters.\n";
    } elsif ($delt == 2) {
      # UTF-8 mode, doublewidth chars
      print "VT is in UTF-8 mode, with doublewidth characters.\n";
      $doublewidth = 1;
    } elsif ($delt == 3) {
      print "VT is in single-byte mode.\n";
      $doublewidth = 0;
      $singlebyte = 1;
    } else {
      print "VT is in unknown mode.  Assuming ASCII-only.\n";
      $singlebyte = 1;
      $doublewidth = 0;
    }
  }
  print $VT100{rst};
}

sub init_utf8_settings()
{
  my $x = 0;
  my $forced = 0;
  while ($x<=$#ARGV) {
    if ($ARGV[$x] eq '--ascii') {
      $singlebyte = 1;
      $doublewidth = 0;
      delete($ARGV[$x]);
      $forced = 1;
    } elsif ($ARGV[$x] eq '--doublewidth=all') {
      $doublewidth = 0x80;
      $singlebyte = 0;
      delete($ARGV[$x]);
      $forced = 1;
    } elsif ($ARGV[$x] eq '--doublewidth=cjk') {
      $doublewidth = 1;
      $singlebyte = 0;
      $forced = 1;
    } else {
      $x++;
    }
  }
  if (!$forced) {
    probe_utf8_settings();
  }
}

sub getpos()
{
  print $VT100{cpr};
  my $notdone = 1;
  my $buf = '';
  my $state = 0;
  my $y = '';
  my $x = '';
  while ($notdone) {
    my $key = ReadKey(0.5);
    if (!defined($key)) {
      print STDERR $VT100{rst}, "No key read; buffer is ",
        &Yeemp::Client::clean($buf),"\n";
      $err = 3;
      return ();
    }
#    print STDERR "$VT100{rst}Got: ",ord($key),"\n";
    if (!defined($key)) {
      $err = 2;
      return ();
    }
    $buf .= $key;
    if ($state==0) {
      if (substr($VT100{rep0},0,length($buf)) ne $buf) {
        if ($buf =~ /^\Q$VT100{rep0}\E$VT100{repy}/s) {
          $state = 1;  # We're building y.
        } else {
          $err = 1;
          return ();
        }
      }
    } elsif ($state == 1) {
      if ($buf =~ /^\Q$VT100{rep0}\E$VT100{repy}\Q$VT100{rep1}\E/s) {
        $state = 2;
      } elsif ($buf !~ /^\Q$VT100{rep0}\E$VT100{repy}/s) {
        $err = 1;
        return ();
      }
    } elsif ($state == 2) {
      if ($buf !~ /^\Q$VT100{rep0}\E$VT100{repy}\Q$VT100{rep1}\E$VT100{repx}$/s) {
        if ($buf =~ /^\Q$VT100{rep0}\E$VT100{repy}\Q$VT100{rep1}\E$VT100{repx}\Q$VT100{rep2}\E$/) {
          return ($2, $1);
        } else {
          $err = 1;
          return ();
        }
      }
    }
  }
}


%control_keys = GetControlChars;
#print STDERR %control_keys;

if (defined($colour_init)) {
  print $colour_init;   # Switch to UTF-8 mode
}

BEGIN
{
  if ($ENV{TERM} eq 'linux') {
    $colour_hotkey="\033[36;1m";	  # Commands
    $colour_text  ="\033[0;32m";		# Normal text
    $colour_prompt="\033[0;35m";  	# Prompts
    $colour_input ="\033[0;36m";	  # Input
    $colour_error ="\033[1;31m";	  # Error messages
    $colour_reset ="\033]R\033c\033\%G";	  # Colour to leave shell at when quit
    $colour_init  ="\033\%G";       # Initialization.  Enable UTF-8.
    $colour_savestate = "\x1b7";
    $colour_restorestate = "\x1b8";
    if (!defined($Yeemp::Config::cycle_palette)) {
      $Yeemp::Config::cycle_palette=1;
    }
    @palette=(
		'000000','990000','009900','999900',
		'000099','990099','009999','999999',
		'666666','ff6666','66ff66','ffff66',
		'6666ff','ff66ff','66ffff','ffffff'
		);
    @origpalette=@palette;
  } else {
    $colour_hotkey='';
    $colour_text='';
    $colour_prompt='';
    $colour_input='';
    $colour_error='';
    $colour_reset='';
    @palette=(
      '000000','990000','009900','999900',
      '000099','990099','009999','999999',
      '666666','ff6666','66ff66','ffff66',
      '6666ff','ff66ff','66ffff','ffffff'
    );
    @origpalette=@palette;
  }
}

sub split_rgb()
{
  return (substr($_[0],0,2),substr($_[0],2,2),substr($_[0],4,2));
}

sub statchange_beep()
{
  print chr(7);
}

sub constrain()
{
  local $val=hex($_[0]);
  local $lim=hex($_[1]);
  if ($val>$lim+0x40) {
    $val=$lim+0x40;
  } elsif ($val<$lim-0x40) {
    $val=$lim-0x40;
  }
  if ($val<0) {
    $val='00';
  } elsif ($val>255) {
    $val&=0xff;
  }
  return sprintf("%02x",$val);
}

sub read_line()
{
  # 0: prompt
  # 1: OK (discarded in text mode)
  # 2: callback
  # 3: Default (optional)
  # 4: Data (optional)
  print "$colour_prompt$_[0]";
  if (defined($_[3])) {
    print "\n(Default: $colour_input$_[3]$colour_prompt): ";
  }
  print $colour_input;
  $line=read_line_utf8();
  if (($line eq '') && defined($_[3])) {
    $line = $_[3];
  }
  &{$_[2]}($line, $_[4]);
}

sub read_function_key($)
{
  my $str = $_[0];
  my $ch;
  while (exists($funckeys{$str})) {
    $ch = ReadKey($function_key_timeout);
    if (defined($ch)) {
      $str .= $ch;
    }
    if (defined($funckeys{$str})) { # Should return Esc after 1/10 sec.
      if (exists($keytrans{$funckeys{$str}})) {
        return ($keytrans{$funckeys{$str}}, $funckeys{$str});
      } else {
        return ("\0", $funckeys{$str});
      }
    }
  }
  $read_char_scratch = $ch;
  if (defined($funckeys{$str})) {
    if (exists($keytrans{$funckeys{$str}})) {
      return ($keytrans{$funckeys{$str}}, $funckeys{$str});
    } else {
      return ("\0", $funckeys{$str});
    }
  } else {
    return ("\0", "");
  }
}

$read_char_scratch = '';

sub read_char_utf8()
{
  # Yeemp Private Use Characters, draft.  They may hop around.
  # 0xE000: UTF-8 tail received sans head.
  # 0xE001: UTF-8 incomplete.
  # 0xE002: UTF-8 overlong.
  # 0xE003: Outside the BMP, skipping.  We'll implement long chars when
  #         someone actually shows me something that uses them.
  #         Heiroglyphic terminals, anyone?
  # 0xE004: Invalid function key.
  no utf8;  # We need to mess with bytes
  my $ch;
  if (!length($read_char_scratch)) {
    $ch = ReadKey(0);
  } else {
    $ch = $read_char_scratch;
    $read_char_scratch = '';
  }
  if (exists($funckeys{$ch})) {
    # $key_pressed is a global used to access depressed function keys OOB.
    ($ch, $key_pressed) = read_function_key($ch);
  } elsif (ord($ch)>=0x80) {
    if ($singlebyte) {
      print "\007";
      return "\0";
    }
    if (ord($ch) < 0xC0) {
      return pack('U', 0xE000);
    } elsif (ord($ch) < 0xE0) { # Two-byte code.
      my $tail = ReadKey(0);
      if ((ord($tail) & 0xC0) != 0x80) {
        return pack('U', 0xE001);
      }
      $ch = ((ord($ch) & 0x1F) << 6) + (ord($tail) & 0x3F);
      if ($ch >= 0x80) {
        return pack('U', $ch);
      }
      return pack('U', 0xE002);
    } elsif (ord($ch) < 0xF0) {
      my $tail = ReadKey(0);
      if ((ord($tail) & 0xC0) != 0x80) {
        return pack('U', 0xE001);
      }
      $ch = ((ord($ch) & 0x0F) << 12) + ((ord($tail) & 0x3F) << 6);
      $tail = ReadKey(0);
      if ((ord($tail) & 0xC0) != 0x80) {
        return pack('U', 0xE001);
      }
      $ch += ord($tail) & 0x3F;
      if ($ch >= 0x800) {
        return pack('U', $ch);
      }
      return pack('U', 0xE002);
    } else {
      return pack('U', 0xE003);
    }
  }
  return $ch;
}

sub doublewidth($)
{
  if ($doublewidth) {
    if ($doublewidth == 0x80) {
      if (ord($_[0])>=0x80) {
        return 1;
      }
    } else {
      my $o = ord($_[0]);
      if ( (($o>=0x900) && ($o<=0x9FF)) # Devanagari
        || (($o>=0x1100) && ($o<=0x137F)) # Hangul, Ethiopic
        # ??? Something wide in 1580-159F
        || (($o>=0x1680) && ($o<=0x169F)) # Ogham
        # Some of the 0x2020 chars are represented as doublewidth?
        || (($o>=0x2400) && ($o<=0x24FF)) # Control codes, circled alphanumerics
        || (($o>=0x2700) && ($o<=0x27FF)) # Dingbats
        || (($o>=0x2E80) && ($o<=0x9FBF)) # Katakana, Kanji
        || (($o>=0xAC00) && ($o<=0xD7BF)) # ???
        || (($o>=0xF900) && ($o<=0xFA3F)) # ???
        # 0xFB30+ has some wide Hebrew
        # 0xFE20-0xFE7F is mixed wide and narrow punctuation
        || (($o>=0xFF00) && ($o<=0xFF5F)) # Wide ASCII
        || (($o>=0xFFE0) && ($o<=0xFFE6)) # Wide currency
        ) {
        return 1;
      }
    }
  } else {
    return 0;
  }
}

sub read_line_utf8()
{
  my $line = '';
  if (defined($colour_savestate)) {
    print $colour_savestate;
  }
  while (1) {
    my $ch = read_char_utf8();
    if ($ch eq $control_keys{ERASE}) {
      if (length($line)) {
        my $bsd = chop($line);
        if (doublewidth($bsd)) {
          print "\x08 \x08";
        }
        print "\x08 \x08";
      }
    } elsif (ord($ch)>=0x20) {
      $line .= $ch;
      print $ch;
    } elsif ($ch eq "\x0a" || $ch eq "\x0d") {
      print "\n";
      return $line;
    } elsif ($ch eq $control_keys{KILL}) {
      if (defined($colour_restorestate)) {
        print $colour_restorestate, ' ' x length($line), $colour_restorestate;
        $line = '';
      }
    } elsif ($ch eq $control_keys{INTERRUPT}) {
      kill(INT, $$);
    } elsif ($ch eq $control_keys{SUSPEND}) {
      stderr("Suspension not yet enabled.\n");
   }
  }
}

sub passwordprompt($$@)
{
  my $prompt=shift;
  my $stty=1;
  if (eval('use IO::Stty; 1;')) {
    IO::Stty::stty(\*STDIN, '-echo');
  } else {
    stderr("Unable to load IO::Stty.  Password may show on screen.\n");
    $stty=0;
  }
  print "$colour_prompt$prompt: $colour_input";
  if (!$stty) {
    print "\033[30m";
  }
  my $line = <STDIN>;
  chomp($line);
  if ($stty) {
    IO::Stty::stty(\*STDIN, 'echo');
  } else {
    print "\033[0m";
  }
  my $sub = shift;
  &{$sub}($line,@_);
}

sub constrain_palette()
{
  local ($r,$g,$b)=&split_rgb($palette[$_[0]]);
  local ($r0,$g0,$b0)=&split_rgb($origpalette[$_[0]]);
  $r=&constrain($r,$r0);
  $g=&constrain($g,$g0);
  $b=&constrain($b,$b0);
#  print "Debug: Constrained $x $palette[$_[0]] against $origpalette[$_[0]] to $r$g$b\n";
  $palette[$_[0]]="$r$g$b";
}

sub cycle_colour()
{
  #     &stderr("Debug: ".((oct('0x'.substr($palette[$x],0,2))+int(rand(32)-16))&0xff)."\n");
  local $x=$_[0];
  $palette[$x]=sprintf("%02x%02x%02x",
    (oct('0x'.substr($palette[$x],0,2))+int(rand(32)-16))&0xff,
    (oct('0x'.substr($palette[$x],2,2))+int(rand(32)-16))&0xff,
    (oct('0x'.substr($palette[$x],4,2))+int(rand(32)-16))&0xff);
  &constrain_palette($x);
  printf ("\033]P%x$palette[$x]",$x);
}

sub cycle_palette()
{
  if ($Yeemp::Config::cycle_palette) {
    if ($Yeemp::Config::cycle_palette==1) {
      local $x=int(rand(16));
      if (rand(7)>1) {
        &cycle_colour($x);
      } else {
        $palette[$x]=$origpalette[$x];
        printf ("\033]P%x$palette[$x]",$x);
      }
    } else {
      local $x=0;
      while ($x<0x10) {
        &cycle_colour($x);
        $x++;
      }
    }
  }
}

sub stderr(@)
{
  if (!defined($colour_error)) {
    $colour_error='[ $colour_error undefined ]';
  }
  print STDERR $colour_error,@_;
}

sub stdout(@)
{
  cycle_palette();
  print $colour_text,@_;
}

sub debug()
{
  if ($Yeemp::Config::debug) {
    print STDERR "Debug: ",@_;
  }
}

sub show_message()
{
  # Current args:
  # 0: Reference to hast containing the following elements:
  # 	userheaders:	User-supplied headers (in the message)
  #		body:			Message body
  #		serverheaders:	Server-supplied headers
  #		raw:			Raw message
  #		decrypterrors:	Decryption errors
  if (!exists(${$_[0]}{old})) {
    print "\007";  # Beep!
    $last_message=$_[0];
    ${$last_message}{old}=1;
  }
  local %head = %{ ${$_[0]}{userheaders} };
  print $colour_input;
  if (defined($head{From})) {
    print &Yeemp::Client::clean("Message from $head{From}: \n");
    delete($head{From});
  } else {
    print "Message from unknown user: \n";
  }
  if (defined(${$_[0]}{decrypterrors}) && (${$_[0]}{decrypterrors} ne '')) {
    print "Decryption errors: ${$_[0]}{decrypterrors}\n";
  }
  local $key;
  foreach $key (keys(%head)) {
    print &Yeemp::Client::clean("$key: $head{$key}\n");
  }
  print &Yeemp::Client::clean("${$_[0]}{body}\n");
}

sub show_statchange()
{
  &stdout($colour_text,&Yeemp::Client::clean("$_[0] changed their status to $_[1]\n"));
}

sub endis()
{
  if ($_[0]) {
    return "enabled";
  }
  return "disabled";
}

sub show_announce_status()
{
  print "${colour_text}Your status will ";
  if (!$Yeemp::Config::announce_status) {
    print "not ";
  }
  print "be announced when you start the client.\n";
}

sub show_settings_window()
{
  print "${colour_text}Your address is $Yeemp::Config::my_address.\n",
    "Status caching is ", &endis($Yeemp::Config::cache_status),
    ".\nLifetime of status change messages is ",
    $Yeemp::Config::status_change_expiry, " seconds.\n";
  &show_announce_status();
  print "Palette cycling is ";
  if ($Yeemp::Config::cycle_palette) {
    if ($Yeemp::Config::cycle_palette==2) {
      print "in tripping mode";
    } else {
      print "in stoned mode";
    }
  } else {
    print "in sober mode";
  }
  print ".\n";
  if ($Yeemp::Config::default_presence_mode eq 'show') {
    print "Your status is shown to all contacts unless you're explicitly hiding from them.\n";
  } elsif ($Yeemp::Config::default_presence_mode eq 'confirm') {
    print "You will be asked whether or not to show your status each time a contact\nrequests your status.\n";
  } elsif ($Yeemp::Config::default_presence_mode eq 'hide') {
    print "Your status is hidden unless otherwise specified.\n";
  } else {
    print "Unhandled status setting $Yeemp::Config::default_presence_mode\n";
  }
  if (keys(%Yeemp::Config::offline)) {
    print "The following servers are currently marked offline:\n";
    foreach $key (keys(%Yeemp::Config::offline)) {
      print "	$key\n";
    }
  }
  print "${colour_hotkey}help set$colour_text for settings help.\n";
}

sub read_message($;$)
{
  # Args:
  # $_[0]: To
  # $_[1] (optional): Subto
  my $header='';
  local ($to,$subto)=@_;
  if (!($to=~/~/)) {
    local $k;
    foreach $k (keys(%Yeemp::Client::contacts_shortnames)) {
      if ($Yeemp::Client::contacts_shortnames{$k}=~/^$to$/i) {
        $to=$k;
      }
    }
  }
  if (!($to=~/~/)) {
    print "${colour_text}'$to' not in contact list.\n";
    return undef;
  }
  local $line = '';
  local $message = '';
  print $colour_text."Writing to $colour_input$to$colour_text";
  if (defined($subto)) {
    print "/$colour_input$subto$colour_text";
  }
  print "\nEnter message, $colour_hotkey.$colour_text on a line by itself when done, $colour_hotkey#$colour_text to cancel.\n$colour_input";
  while (defined($line) && (!($line=~/^[.#]\n/s))) {
    $message.=$line;
    $line = read_line_utf8()."\x0a";
  }
  if (!defined($line)) {
    print "${colour_text}EOF on stdin, *NOT* sending message.\n";
    return;
  } elsif ($line eq "#\n") {
    print "${colour_text}Message cancelled.\n";
    return;
  } elsif ($line eq ".\n") {
    print "${colour_text}Delivering.\n";
    local $fullmessage='';
    if (defined($subto)) {
      &UI::debug("Subto: $subto\n");
      $fullmessage.="Subto: $subto\012";
    }
    $fullmessage.="From: $Yeemp::Config::my_address\012\012$message";
    &Yeemp::Client::deliver_message($to, $fullmessage);
  } else {
    print "Invariancy violation in terminal line.  Please reboot universe to continue.\n";
  }
}

sub show_client_help()
{
  if (defined($_[0]) && $_[0] eq 'set') {
    &show_settings_help();
    return;
  }
  print "${colour_text}Commands:\n";
  print "${colour_hotkey}a$colour_text: show current message again.\n";
	print "${colour_hotkey}add [address]$colour_text: add contact.\n";
  print "${colour_hotkey}credits$colour_text: show credits.\n";
  print "${colour_hotkey}edit$colour_text: edit a contact.\n";
  print "${colour_hotkey}get$colour_text: request a file.\n";
  print "${colour_hotkey}installkey$colour_text: install your public key.\n";
  print "${colour_hotkey}msg$colour_text: send message.\n";
  print "${colour_hotkey}n$colour_text: next message.\n";
  print "${colour_hotkey}r$colour_text: reply.\n";
  print "${colour_hotkey}raw$colour_text: show raw message.\n";
  print "${colour_hotkey}send$colour_text: send file.\n";
  print "${colour_hotkey}settings$colour_text: show settings.\n";
  print "${colour_hotkey}status$colour_text: set your status.\n";
  print "${colour_hotkey}w$colour_text: who's online.\n";
  print "${colour_hotkey}q$colour_text: quit\n";
}

sub show_settings_help()
{
  print "${colour_text}All configuration commands begin with ${colour_hotkey}set\n";
  print "set offline ${colour_input}host$colour_text: Mark host as offline.\n";
  print "${colour_hotkey}set online$colour_text: Mark all offline hosts as online.\n";
  print "${colour_hotkey}set presence$colour_text: Configure status publicization.\n";
  print "${colour_hotkey}set announce on$colour_text: Announce your status when you connect.\n";
  print "${colour_hotkey}set announce off$colour_text: or don't, if you prefer.\n";
  print "${colour_hotkey}set sober$colour_text: Disable palette shifting.\n";
  print "${colour_hotkey}set stoned$colour_text: Mild palette shifting.\n";
  print "${colour_hotkey}set tripping$colour_text: Strong palette shifting.\n";
}

sub get_file_prompt()
{
  # Args: 0: Tail of 'get' command.
  local ($who,$file)=&Yeemp::Client::split_filespec($_[0]);
  if (!defined($who)) {
    print "Get file from who? ";
    $who=read_line_utf8();
    if ($who!~/~/) {
      $who=Yeemp::Client::shortname_to_address($who);
    }
  }
  if (!defined($who)) {
    print "Nevermind.\n";
    return;
  }
  if (!defined($file)) {
    print "File to get: ";
    $file=read_line_utf8();
    if ($file eq '') {
      print "Nevermind.\n";
      return;
    }
    if ($file!~/^\//) {
      $file="/$file";
    }
  }
  &Yeemp::Client::deliver_message($who,"From: $Yeemp::Config::my_address\012Get: $file\012\012");
}

sub show_who_listing()
{
  local $k;
  foreach $k (keys(%Yeemp::Client::contacts)) {
    local $stat='';
    if (exists($Yeemp::Client::contacts_status{$k})) {
      $stat=$Yeemp::Client::contacts_status{$k};
    }
    if (exists($Yeemp::Client::contacts_shortnames{$k})) {
      $k=$Yeemp::Client::contacts_shortnames{$k};
    }
    print sprintf("$colour_text\%-40s$colour_input\%-39s\n",$k,$stat);
  }
}

sub send_file_prompt()
{
  # arg 0: Argument tail.
  local ($who,$file)=&Yeemp::Client::split_filespec($_[0]);
  if (!defined($who)) {
    print "${colour_prompt}Send file to whom? $colour_input";
    $who=read_line_utf8();
    if ($who!~/~/) {
      $who=&Yeemp::Client::shortname_to_address($who);
    }
  }
  if (!defined($who)) {
    print "${colour_text}Nevermind.\n";
    return;
  }
  if (!defined($file)) {
    print "${colour_prompt}File to send: $colour_input";
    $file=read_line_utf8();
    if ($file eq '') {
      print "${colour_text}Nevermind.\n";
      return;
    }
#    if ($file!~/^\//) {
#      $file="/$file";
#    }
  }
  &Yeemp::Client::send_file($who,$file);
}

sub set_presence()
{
  local $notdone=1;
  while ($notdone) {
    print "${colour_prompt}Select default presence mode: \n";
    print "${colour_hotkey}hide$colour_prompt, ${colour_hotkey}show$colour_prompt, or ${colour_hotkey}confirm$colour_prompt: ";
    $line=read_line_utf8();
    $notdone=0;
    if ($line=~/^c/i) {
      $Yeemp::Config::default_presence_mode='confirm';
    } elsif ($line=~/^h/i) {
      $Yeemp::Config::default_presence_mode='hide';
    } elsif ($line=~/^s/i) {
      $Yeemp::Config::default_presence_mode='show';
    } else {
      print "${colour_text}Unsupported presence mode\n";
    }
  }
}

sub set_setting()
{
  local $setting=$_[0];
  $setting=~s/^ //gs;
  local @setting=split(/ /,$setting);
  if ($setting[0]=~/^anno/i) {
    if ($#setting==1) {
      if ($setting[1]=~/^off$/i) {
        $Yeemp::Config::announce_status=0;
      } elsif ($setting[1]=~/^on$/i) {
        $Yeemp::Config::announce_status=1;
      } else {
        &UI::stderr("I don't understand $setting[1].  Try 'on' or 'off'.\n");
      }
    }
    &show_announce_status();
  } elsif ($setting=~/^off(.*?) ([^ ](.*?))$/i) {
    local $host=$2;
    print "$host set offline.\n";
    $Yeemp::Config::offline{$host}=1;
  } elsif ($setting=~/^online/i) {
    undef(%Yeemp::Config::offline);
    print "Offline host cache cleared.\n";
  } elsif ($setting=~/^pres/i) {
    &set_presence();
  } elsif ($setting=~/^sober/i) {
    $Yeemp::Config::cycle_palette=0;
    print "System sobered up. [Palette reset sequence: \033]R]\n";
  } elsif ($setting=~/^ston/i) {
    $Yeemp::Config::cycle_palette=1;
    print "${colour_text}Inhaling.\n";
  } elsif ($setting=~/^trip/i) {
    $Yeemp::Config::cycle_palette=2;
    &cycle_palette();
    print "${colour_text}Oooh.  The colours.\n";
  }
}

sub show_credits()
{
  print "$colour_text\n";
  print "Yeemp version 0.4.5\n";
  print "Copyright 2001 Nathaniel E. Bray/Deekoo L.\n";
  print "Licensed under the GPL and the Net::SSLeay Documentation License.\n";
  print "\n";
  print "This product includes software developed by Sampo Kellomaki (sampo\@iki.fi)\n";
  print "\n";
}

sub nomess_prompt()
{
  # args: 0: Command line; 1: Message hashref.
  local $line=$_[0];
  if ($line=~/^a$/i) {
    &show_message($last_message);
	} elsif ($line=~/^add/i) {
	  if ($line=~/^add[ ]+(.*)$/i) {
      local $address=$1;
			$Yeemp::Client::contacts_shortnames{$address}='';
      &Yeemp::Client::edit_contact($address);
		} else {
		  &stderr("You need to specify an address.\n");
		}
  } elsif ($line=~/^cred/i) {
    &show_credits();
  } elsif ($line =~/^edit[ ]+(.*)$/i) {
    local $cont=$1;
    while ($cont=~/ $/s) {
      $cont=~s/ $//gs;
    }
    &Yeemp::Client::edit_contact($cont);
  } elsif ($line=~/^get(.*?)$/i) {
    local $tail=$1;
    $tail=~s/^ //gs;
    &get_file_prompt($tail);
  } elsif ($line=~/^installkey$/i) {
    &Yeemp::Client::copy_login_key($Yeemp::Config::server_name,
      $Yeemp::Config::my_address);
  } elsif ($line=~/^msg(| (.*?))$/) {
    local $to=$2;
    if ((!defined($to)) || ($to eq '')) {
      print "${colour_prompt}To: $colour_input";
      $to=&read_line_utf8();
    }
    local $subto;
    if ($to=~/^(.*?)\/(.*?)$/s) {
      $to=$1;
      $subto=$2;
    }
    &read_message($to,$subto);
  } elsif ($line =~ /^q(|uit)$/i) {
    &Yeemp::Client::shutdown_client();
    exit;
  } elsif ($line =~ /^r$/i) {
    if (defined($Yeemp::Client::last_from)) {
      $replyto=$Yeemp::Client::last_from;
    } else {
      print "${colour_text}Message has no from.\n";
      print "${colour_prompt}Send reply to: ${colour_input}";
      $replyto=read_line_utf8();
    }
    &read_message($replyto, $Yeemp::Client::last_subfrom);
  } elsif ($line=~/^raw(.*?)$/i) {
    print "${colour_text}Raw message:\n$colour_message";
    print &Yeemp::Client::clean($Yeemp::Client::last_raw);
  } elsif ($line=~/^send(.*?)$/i) {
    local $tail=$1;
    &send_file_prompt($tail);
  } elsif ($line =~/^set(|tings)(| (.*?))$/i) {
    local $opt=$3;
    if (defined($opt)) {
      &set_setting($opt);
    } else {
      &show_settings_window();
    }
  } elsif ($line=~/^status(| (.*?))$/i) {
    local $status=$2;
    if (defined($status)) {
      $status=~s/^ //gs;
      &Yeemp::Client::update_status($status);
    } else {
      print "${colour_text}Current status: $colour_input$Yeemp::Client::my_status\n";
    }
  } elsif ($line=~/^w$/i) {
    &show_who_listing();
  } elsif (($line eq '?') || ($line =~/^help(| (.*?))$/i) ||
    ($line =~ /^\/help$/i)) {
    local $arg=$2;
    &show_client_help($arg);
  } elsif ($line ne '') {
    $line=~s/[\000-\037]/?/gs;
    print "${colour_text}I don't understand how to $line.  Type ${colour_hotkey}?${colour_text} for help.\n";
  }
}

sub prompt_nomess()
{
  select(STDOUT); $|=1;
  print "${colour_prompt}emp> $colour_input";
  local $line=&read_line_or_check();
  if (!defined($line)) {
    return;
  }
  &nomess_prompt($line);
}

sub read_line_or_check()
{
  local $line;
  while (1) {
    local $rin='';
    local $eout;
    select((select(Yeemp::Channel::Channel), $|=1)[0]);
    vec($rin,fileno(STDIN),1)=1;
    if (defined(fileno(Yeemp::Channel::CHANNEL))) {
      vec($rin,fileno(Yeemp::Channel::CHANNEL),1)=1;
    }
    local $selected=select($rin,undef,$eout=$rin,5);
    if (vec($rin,fileno(STDIN),1)) {
      $line=read_line_utf8();
      return $line;
    }
    if (ord($eout)) {
      &UI::debug("Eout: ".ord($eout)."\n");
    }
    if (defined(fileno(Yeemp::Channel::CHANNEL))) {
      if (vec($rin,fileno(Yeemp::Channel::CHANNEL),1)) {
        &Yeemp::Channel::channel_slurp(CHANNEL);
        if (&Yeemp::Client::check()) {
          return undef;
        }
      }
    }
  }
}

sub message_prompt()
{
  # New args:
  # 0: Reference to hash containing message.
  local $notdone=1;
  local $replyto;
  &cycle_palette();
  select(STDOUT);
  $|=1;
  print "${colour_prompt}emp> $colour_input";
  local %headers=%{${$_[0]}{userheaders} };
  while ($notdone) {
    local $line=&read_line_or_check();
    if (!defined($line)) {
      return 1;
    }
    if (defined($line)) {
      $line=~s/^ +//s;
      $line=~s/ +$//s;
      if ($line =~ /^a$/i) {
        return 0;
      } elsif ($line =~ /^d$/i) {
        &UI::stderr("Deleting IM from log not yet supported.\n");
      } else {
        &nomess_prompt($line);
        return 1;
      }
      &cycle_palette();
      print "${colour_prompt}emp> ${colour_input}";
    }
  }
}

sub show_public_key_request()
{
  # Arg0: User
  # Arg1: Requestor's Key
  # Arg2: Key Fingerprint
  print "${colour_text}Public key request from ${colour_input}$_[0]${colour_text}.\n";
  print "(Requesting key fingerprint ${colour_input}$_[2]${colour_text})\n";
  print "${colour_prompt}Grant it (y/n)? $colour_input";
  while (1) {
    local $line=read_line_utf8();
    if ($line=~/^y(|es)$/i) {
      &Yeemp::Client::grant_public_key_request($_[0],$_[1]) or
        print "${colour_error}There was an error granting the public key request.\n";
      return;
    } elsif ($line=~/^[Nn]$/) {
      print "${colour_text}Public key request not granted.\n";
      return;
    } else {
      print "I am not understanding the silliness you are inputting.\n";
      print "Please to input the logical command that is making the sense.\n";
    }
  }
}

sub shutdown_ui()
{
  print $colour_reset;
  ReadMode 0;
}

sub ynprompt($$$;$)
{
  # $_[0]: question
  # $_[1]: yes function
  # $_[2]: data for yes function
  # $_[3]: (optional) No function.  Takes the same data as the yes function.
  local $notdone=1;
  while ($notdone) {
    $notdone=0;
    print "$colour_prompt$_[0]$colour_input";
    local $line=read_line_utf8();
    if ($line=~/^y/i) {
      &{$_[1]}($_[2]);
    } elsif ($line=~/^n/i) {
      if (defined($_[3])) {
        &{$_[3]}($_[2]);
      }
    } else {
      print "Incomprehensible input.\n";
      $notdone=1;
    }
  }
}

sub show_inbound_public_key()
{
  # $_[0]: Sender.
  # $_[1]: Key.
  print "Public key received from $_[0].\n";
  print "(Fingerprint $Yeemp::Client::fingerprint[0])\n";
  print "Use it? ";
  local $line;
  while ($line=read_line_utf8()) {
    if ($line=~/^[Dd]$/) {
      print "Nuking.\n";
      &Yeemp::Client::delete_current_im();
      return 1;
    } elsif ($line=~/^[Nn]$/) {
      print "Good.  I like a suspicious mind.\n";
      return 1;
    } elsif ($line=~/^[Qq]$/) {
      &Yeemp::Client::shutdown_client();
    } elsif ($line=~/^[Yy]$/) {
      print "Silly.  Importing public key...\n";
      &Yeemp::Client::import_contact_pubkey($_[0],$_[1]);
      return 1;
    }
  }
}

sub edit_contact_by_address()
{
  local $notdone=1;
  local $address=$_[0];
  while ($notdone) {
    print "${colour_text}Editing contact list info for: $address\n";
    if (exists($Yeemp::Client::contacts_shortnames{$address})) {
      print "Name: $colour_input$Yeemp::Client::contacts_shortnames{$address}\n";
    } else {
      print "No short name defined.\n";
    }
    print "${colour_text}Presence authorization mode: ";
    if (exists($Yeemp::Client::contacts_presence{$address})) {
      print "$colour_input$Yeemp::Client::contacts_presence{$address}\n";
    } else {
      print "default ($colour_input$Yeemp::Config::default_presence_mode$colour_text)\n";
    }
#    print "Require signatures (unimplimented): \n";
    print "${colour_prompt}edit>$colour_input ";
    local $line=read_line_utf8();
    if ($line =~/^auth/) {
      local $badmode=0;
      if ($line=~/[ ]+(.*?)$/) {
        local $mode=$1;
        if ($mode =~ /^default$/i) {
          delete($Yeemp::Client::contacts_presence{$address});
        } elsif (($mode eq 'confirm') || ($mode eq 'show') ||
          ($mode eq 'hide')) {
          $Yeemp::Client::contacts_presence{$address}=$mode;
        } else {
          $badmode=1;
        }
      } else {
        $badmode=1;
      }
      if ($badmode) {
        print "${colour_text}Presence authorization determines whether or not to display your\n";
        print "status to $address.  There are four modes.\n";
        print "${colour_input}show$colour_text will enable them to see\n";
        print "your status automatically when you start Yeemp or when they\n";
        print "request your status.  Yeemp will automatically request the status\nof users set to show mode.\n";
        print "${colour_input}confirm$colour_text will ask your permission to show them your\n";
        print "status each you receive a status request from them.\n";
        print "${colour_input}hide$colour_text will never show them your status unless you manually send it to them.\n";
        print "${colour_input}default$colour_text will use whichever of the above modes is your default.\n";
      }
    } elsif ($line =~/^name/) {
      if ($line=~/[ ]+(.*)$/) {
        $Yeemp::Client::contacts_shortnames{$address}=$1;
      } else {
        print "${colour_prompt}New name for $address> $colour_input";
        $Yeemp::Client::contacts_shortnames{$address}=&read_line_utf8();
      }
    } elsif ($line =~ /^purge$/i) {
      &ynprompt("This will erase all keys associated with this client, and all unsent messages\nto them.  Are you absolutely certain you want to purge this contact? ", \&Yeemp::Client::purge_contact, $address);
      return;
    } elsif ($line =~ /^q/i) {
      $notdone=0;
    } else {
      print "${colour_text}Edit commands:\n";
      print "${colour_hotkey}auth$colour_text: Change presence authorization mode.\n";
      print "${colour_hotkey}purge$colour_text: Purge this contact.\n";
      print "${colour_hotkey}name$colour_text: Change contact's short name.\n";
      print "${colour_hotkey}q$colour_text: Quit to main.\n";
    }
  }
}

sub init_preferences()
{
  local $notdone=1;
  while ($notdone) {
    if (!defined($Yeemp::Config::server_name)) {
      print "${colour_prompt}Server name: $colour_input";
      $Yeemp::Config::server_name=read_line_utf8();
    }
    if (!defined($Yeemp::Config::my_address)) {
      print "${colour_prompt}Your Yeemp address: $colour_input";
      $Yeemp::Config::my_address=read_line_utf8();
      if ($Yeemp::Config::my_address !~ /~/) {
        stderr("You don't have a ~ in your address.  Your address should be user~machine,\nwhere machine is your Yeemp server.  Otherwise, you will not be able to access\nyour account.");
      }
    }
    print "${colour_text}Your settings:\nServer: $colour_input$Yeemp::Config::server_name\n";
    print "${colour_text}Address: $colour_input$Yeemp::Config::my_address\n";
    print "${colour_prompt}Is this correct? $colour_input";
    local $line=read_line_utf8();
    if ($line=~/^y/i) {
      $notdone=0;
    } else {
      undef($Yeemp::Config::my_address);
      undef($Yeemp::Config::server_name);
    }
  }
}

sub show_image()
{
  &stdout("$_[1]\n");
  &stderr("Unable to show image from $_[0] in console mode.\n");
}

1;
