package UI;
use Gtk;
use Gtk::Gdk;
use Gtk::Keysyms;
use Yeemp::UI::GtkUTF8;
use POSIX qw(setlocale);

sub cb_noset_lang()
{
  Gtk->_exit(0);
}

sub cb_set_lang()
{
  $ENV{LANG} = $ENV{LANG}.'.UTF-8';
  Gtk->_exit(1);
}

sub load_locale()
{
  use Yeemp::UI::Locale::C;
  %locale = %UI::Locale::C::locale;
  if (exists($ENV{LANG})) {
    if ($ENV{LANG}=~/^([a-z][a-z])/) {
      my $lang = $1;
      if (!eval("use Yeemp::UI::Locale::$lang; \%locale = (\%locale, \%UI::Locale::${lang}::locale); 1;")) {
        stderr("Unable to load language $lang.\n");
      }
    }
    if ($ENV{LANG}!~/\.UTF-8$/) {
      my $pid = fork();
      if ($pid) {
        waitpid($pid,0);
        if ($?==0x100) {
          setlocale(&POSIX::LC_CTYPE,$ENV{LANG}.'UTF-8');
          if (setlocale(&POSIX::LC_CTYPE) ne "$ENV{LANG}.UTF-8") {
            stderr("Invalid locale $ENV{LANG}.UTF-8.  Trying en_US.UTF-8.\n");
            setlocale(&POSIX::LC_CTYPE,'en_US.UTF-8');
          }
        }
      } else {
        init Gtk;
        ynprompt("LANG environment variable is currently $ENV{LANG}.  Set it to $ENV{LANG}.UTF-8?",
          \&cb_set_lang, undef, \&cb_noset_lang);
        main Gtk;
      }
    }
  } else {
    stderr("LANG environment variable not set.  Assuming ASCII/C, which means that only\nLatin and Swahili may be properly displayed (along with most, but not all, of\nEnglish.)\n");
  }
}

BEGIN
{
	local $^W=0;
	%presence_string =
    (
      'hide' => "Never show presence",
      'confirm' => "Ask each time",
      'show' => "Always show presence",
      undef, "Default");
																																					
  return 1;
}

sub stderr(@)
{
  local $string=join('',@_);
  if (defined($$stderr_textfield)) {
    $$stderr_textfield->insert_text($string,-1);
    $$stderr_textfield->show;
    # //FIXME: Add redraw code.
  }
  $stderr_text.=$string;
  &stdout(@_);
  print STDERR @_;
}

sub debug(@)
{
  if ($Yeemp::Config::debug) {
    stderr("Debug: ",@_);
  }
}

sub stdout(@)
{
  local $string=join('',@_);
  # //FIXME: Add code to trim messages out of box when it becomes sufficiently
  # large to warrant that.
  $stdout_text.=$string;
  if (defined($$stdout_textfield)) {
    $$stdout_textfield->insert_text($string,-1);
    $$stdout_textfield->show;
    # //FIXME: Add redraw code
  }
  if (defined($statline)) {
    $statline->set($string);
  } else {
    print STDERR "Error in stdout: statline undef.\n";
  }
}

sub statchange_beep()
{
  # XXX Beep for X unwritten
  stderr("X statchange_beep unwritten.\n");
}

sub icon($)
{
  if (-f $Yeemp::Client::clientdir."icons/$_[0].xpm") {
    return $Yeemp::Client::clientdir."icons/$_[0].xpm";
  }
  my $home = (getpwuid($<))[7];
  if (-f "$home/Yeemp/icons/$_[0].xpm") {
    return "$home/Yeemp/icons/$_[0].xpm";
  }
  if (-f "/usr/local/share/yeemp/icons/$_[0].xpm") {
    return "/usr/local/share/yeemp/icons/$_[0].xpm";
  }
  if (-f "/usr/share/yeemp/icons/$_[0].xpm") {
    return "/usr/share/yeemp/icons/$_[0].xpm";
  }
  return undef;
}

sub pw_clear($)
{
  ${$_[0]}{string} = '';
  ${$_[0]}{'pos'} = 0;
  ${$_[0]}{state} = 0;
  my @butts = ${$_[0]}{hbox}->children;
  for my $b (@butts) {
    my $nb = new Gtk::Pixmap(${$_[0]}{buttup}, ${$_[0]}{buttup_m});
    $b->widget->destroy;
    ${$_[0]}{hbox}->pack_start($nb, 0, 0, 0);
    $nb->show;
  }
}

sub pw_appendchar($%$)
{
  my ($pwobj, $keyval) = (@_);
  my @butts = ${$pwobj}{hbox}->children();
  $butts[$#butts]->widget->destroy;
  if (${$pwobj}{state}==0) {
    my $down = new Gtk::Pixmap(${$pwobj}{buttdown}, ${$pwobj}{buttdown_m});
    ${$pwobj}{hbox}->pack_start($down, 0, 0, 0);
    ${$pwobj}{hbox}->reorder_child($down, 0);
    $down->show;
  } else {
    my $up = new Gtk::Pixmap(${$pwobj}{buttup}, ${$pwobj}{buttup_m});
    ${$pwobj}{hbox}->pack_start($up, 0, 0, 0);
    ${$pwobj}{hbox}->reorder_child($up, 0);
    $up->show;
  }
  ${$pwobj}{string}.=chr($keyval);
  ${$pwobj}{'pos'}++;
  if (${$pwobj}{'pos'} >= 8) {
    ${$pwobj}{'pos'} = 0;
    if (${$pwobj}{state}) {
      ${$pwobj}{state} = 0;
    } else {
      ${$pwobj}{state} = 1;
    }
  }
}

sub pw_enter($%)
{
  &{${$_[0]}{callback}}(${$_[0]}{string}, @{${$_[0]}{args}});
  ${$_[0]}{win}->destroy;
}

sub pw_backspace($%)
{
  my $p = $_[0];
  my @butts = ${$p}{hbox}->children;
  if (${$p}{'pos'}==0) {
    ${$p}{'pos'}=8;
    if (${$p}{state}) {
      ${$p}{state}=0;
    } else {
      ${$p}{state}=1;
    }
  }
  substr(${$p}{string},-1,1) = '';
  $butts[${$p}{'pos'}-1]->widget->destroy;
  my $butt='up';
  if (${$p}{state}) { $butt='down'; }
  my $wid = new Gtk::Pixmap(${$p}{"butt$butt"}, ${$p}{"butt${butt}_m"});
  ${$p}{hbox}->pack_start($wid, 0, 0, 0);
  ${$p}{hbox}->reorder_child($wid,${$p}{'pos'});
  ${$p}{'pos'}--;
  $wid->show;
}

sub cb_pw_keypress($$$)
{
  # Hashref, Hashref, Evt
  my %e = %{pop(@_)};
  my $pwobj=$_[1];
  my $kv = $e{'keyval'};
  if ($kv ==  $Gtk::Keysyms{Control_L}) {
    ${$pwobj}{Control_L} = 1;
  } elsif ($kv == $Gtk::Keysyms{Control_R}) {
    ${$pwobj}{Control_R} = 1;
  } elsif ($kv == $Gtk::Keysyms{BackSpace}) {
    pw_backspace($pwobj);
  } elsif (($kv == $Gtk::Keysyms{Linefeed}) || ($kv == $Gtk::Keysyms{Return})) {
    pw_enter($pwobj);
  } elsif (($kv == $Gtk::Keysyms{'U'}) || ($kv == $Gtk::Keysyms{'u'})) {
    if (${$pwobj}{Control_L} || ${$pwobj}{Control_R}) {
      pw_clear($pwobj);
    } else {
      pw_appendchar($pwobj, $kv);
    }
  } elsif ($kv<0xF000) {
    pw_appendchar($pwobj, $kv);
  }
  return 0;
}

sub cb_pw_keyrel($$)
{
  my %e=%{pop(@_)};
  if ($e{keyval} == $Gtk::Keysyms{Control_L}) {
    ${$_[0]}{Control_L} = 0;
  } elsif ($e{keyval} == $Gtk::Keysyms{Control_R}) {
    ${$_[0]}{Control_R} = 0;
  }
  return 0;
}

sub passwordprompt($$@)
{
  # Args:
  # 0: Prompt
  # 1: Function to call w/PW.
  # 2..$#_: Extra args for func
  # Func will be called with PW as zeroeth arg.
  my %pwobj = ('pos' => 0, string => '', state => 0);
  $pwobj{win} = new Gtk::Window;
  $pwobj{win}->set_events(['key_press_mask', 'key_release_mask']);
  my $vbox = new Gtk::VBox(0,1);
  my $label = new Gtk::Label(shift);
  $vbox->pack_start($label,0,0,0);
  $label->show;
  $pwobj{win}->signal_connect('key_press_event', \&cb_pw_keypress, \%pwobj);
  $pwobj{win}->signal_connect('key_release_event', \&cb_pw_keyrel, \%pwobj);
  $pwobj{hbox} = new Gtk::HBox(0,1);
  $pwobj{callback} = shift;
  $pwobj{args} = \@_;
  $pwobj{win}->realize;
  ($pwobj{'buttdown'}, $pwobj{'buttdown_m'}) = 
    Gtk::Gdk::Pixmap->create_from_xpm($pwobj{win}->window,
      Gtk::Gdk::Color->parse_color('black'), icon('buttdown'));
  ($pwobj{'buttup'}, $pwobj{'buttup_m'}) =
    Gtk::Gdk::Pixmap->create_from_xpm($pwobj{win}->window, 
    Gtk::Gdk::Color->parse_color('black'), icon('buttup'));
  $vbox->pack_start($pwobj{hbox}, 0, 0, 0);
  my $x=0;
  while ($x<8) {
    my $up = new Gtk::Pixmap($pwobj{buttup}, $pwobj{buttup_m});
    $up->show;
    $pwobj{hbox}->pack_start($up, 0, 0, 0);
    $x++;
  }
  $pwobj{hbox}->show;
  $ok=new Gtk::Button('Enter Password');
  $ok->signal_connect('clicked', \&pw_enter, \$pwobj);
  $vbox->pack_start($ok,0,0,0);
  $ok->show;
  $vbox->show;
  $pwobj{win}->add($vbox);
  $pwobj{win}->show;
}

sub show_settings_window()
{
  local $win=new Gtk::Window;
#  local $vbox=new Gtk::VBox(0,1);
#  local $label=new Gtk::Label('Your Yeemp settings');
#  local $addrbox=new Gtk::HBox(0,1);
#  local $addrlabel=new Gtk::Label('Address');
#  local $addrent=new Gtk::Label($Yeemp::Config::my_address);

  local $label=new Gtk::Label(
    "Your address is $Yeemp::Config::my_address.\n\n".
    'Status caching is '.
    ($Yeemp::Config::cache_status ? 'Enabled' : 'Disabled').
    "\n\n".
    'Lifetime of status change messages is '.
       $Yeemp::Config::status_change_expiry.
    ' seconds.'
    );
  $win->add($label);
  $label->show;
  $win->show;
}

sub read_line($$$;@)
{
  my $entry = new UI::Entry(@_);
  $entry->show;
}

sub cb_send_file()
{
  &stderr("cb_send_file not yet written.\n");
  0;
}

sub cb_req_file()
{
  &stderr("cb_req_file not yet written.\n");
  0;
}

sub cb_del_contact_confirmed()
{
  delete($Yeemp::Client::contacts{$_[0]});
  delete($Yeemp::Client::contacts_sendfrom{$_[0]});
  delete($Yeemp::Client::contacts_recvfrom{$_[0]});
  delete($Yeemp::Client::contacts_shortnames{$_[0]});
  &Yeemp::Client::save_contact_list();
  &redraw_main_window();
  &UI::debug("del_contact: @_\n");
#  cb_closebutton_3rdgen(\$_[0]);
  0;
}

sub init_bogus_stdio()
{
  $stdout_text='';
  $stderr_text='';
  $statline=new Gtk::Label('');
#  $statline=new Gtk::Label('');
}

sub init()
{
  load_locale();
#  $ENV{LANG}='en_US.UTF-8';
  init Gtk;
  $bgcolour = new Gtk::Gdk::GC(Gtk::Gdk->ROOT_PARENT());
  my $cmap = Gtk::Gdk->ROOT_PARENT->get_colormap;
  $bgcolour->set_foreground($cmap->color_alloc(
    { red => 0x0, green => 0, blue => 0 } ));
  $textcolour = new Gtk::Gdk::GC(Gtk::Gdk->ROOT_PARENT());
  $textcolour->set_foreground($cmap->color_alloc(
    { red => 0x00, green => 0xffff, blue => 0 } ));
  $cursorcolour = new Gtk::Gdk::GC(Gtk::Gdk->ROOT_PARENT());
  $cursorcolour->set_foreground($cmap->color_alloc(
    { red => 0, green => 0xffff, blue => 0xc000 } ));
  $entryfont = Gtk::Gdk::Font->fontset_load(
    '-*-unifont-*-*-*-*-*-*-*-*-*-*-*-*,'.
    '-*-*-*-*-*-*-*-*-*-*-*-*-*-*'
#    '-*-*-*-*-*-*-*-*-*-*-*-*-iso8859-1'
    );
  if (!defined($UI::entryfont)) {
    UI::stderr("Warning: Undef entry font.\n");
    exit(1);
  }
      
  # &create_stream('STDOUT','Log');
  init_bogus_stdio();
  Gtk::Rc->parse_string(
'style "Yeemp_style"
{
 fontset = "-*-unifont-*-*-*-*-*-*-*-*-*-*-iso10646-*"
}
widget "*" style "Yeemp_style"
class "GtkWidget" style "Yeemp_style"
');
  $openwindows=0;
}

sub close_stdout_win()
{
  $stdout_win->destroy();
  undef($stdout_win);
}

sub close_stderr_win()
{
  $stderr_win->destroy();
  undef($stderr_textfield);
  undef($stderr_win);
}

sub show_stderr()
{
  if (defined($stderr_win)) {
    &close_stderr_win();
  }
  $stderr_win=new Gtk::Window;
  local $vbox=new Gtk::VBox;
  local $label=new Gtk::Label('Errors');
  local $hbox=&text_hbox($stderr_text,$stderr_textfield);
  $vbox->pack_start($label,0,0,0);
  $label->show;
  $vbox->pack_start($hbox,0,0,0);
  $vbox->show;
  $stderr_win->add($vbox);
  $stderr_win->show;
}

sub text_hbox()
{
  local $hbox=new Gtk::HBox;
  local $text=new Gtk::Text;
  local $vscroll=new Gtk::VScrollbar($text->vadj);
  $text->insert_text($_[0]);
  if (exists($_[1])) {
    $_[1]=\$text;
#    bless($_[1],Gtk::Text);
  }
  $hbox->pack_start($text,1,0,0);
  $hbox->pack_start($vscroll,0,0,0);
  $text->show;
  $vscroll->show;
  $hbox->show;
  return $hbox;
}

sub show_stdout()
{
  if (defined($stdout_win)) {
    &close_stdout_win();
  }
  $stdout_win=new Gtk::Window;
  local $vbox=new Gtk::VBox;
  local $hbox=&text_hbox($stdout_text,$stdout_textfield);
  local $label=new Gtk::Label('Yeemp action log');
  local $stderrbutt=new Gtk::Button('Show errors only');
  $vbox->pack_start($label,0,0,0);
  $vbox->pack_start($hbox,1,0,0);
  $vbox->pack_start($stderrbutt,0,0,0);
  $stderrbutt->signal_connect('clicked',\&show_stderr);
  $label->show;
  $stderrbutt->show;
  $vbox->show;
  $stdout_win->add($vbox);
  $stdout_win->show;
}

sub draw_main_window()
{
    $main_window = new Gtk::Window;
    local $accel = new Gtk::AccelGroup;
    $accel->attach($main_window);
    local $factory = new Gtk::ItemFactory('Gtk::MenuBar', '<main>', $accel);
    $factory->create_items(
        { path => '/_Yeemp', type => '<Branch>', },
        { path => '/_Yeemp/tearoff', type => '<Tearoff>', },
        { path => '/_Yeemp/Get _New Messages',
			callback => [ \&main::show_new_messages ], 
            accelerator => '<control>N' },
        { path => '/_Yeemp/_Add Contact',
            accelerator => '<control>A', callback => [ \&main::add_contact ], },
        { path => '/_Yeemp/_Set Status',
            accelerator => '<control>S', callback => [ \&main::change_status ], },
#		{ path => '/_Yeemp/_Reconnect to server',
#			accelerator => '<control>R', callback => [ \&Yeemp::Client::reconnect_to_server ], },
		{ path => '/_Yeemp/_Credits',
			accelerator => '<control>C', callback => [ \&show_credits ], },
        { path => '/_Yeemp/_Preferences',
			accelerator => '<control>P', callback => [ \&init_preferences ], },
        { path => '/_Yeemp/_Quit',
            accelerator => '<control>Q', callback => [ \&main::quit_client ], }
    );
    local $menubar = $factory->get_widget('<main>');
    $menubar->show;
    local $main_window_vbox=new Gtk::VBox 0,1;
    $main_window_vbox->pack_start($menubar,0,0,1);
    $contact_list_viewport=new Gtk::Viewport undef,undef;
    $main_window_vbox->pack_start($contact_list_viewport,1,1,1);
    $contactbox=new Gtk::VBox 0,1;
    local $contact;
    foreach $contact (keys(%Yeemp::Client::contacts)) {
        &add_contact_to_box($contactbox,$contact);
    }
    $contact_list_viewport->add($contactbox);
    $contactbox->show;
    $contact_list_viewport->show;
#    &create_stream_widgets('STDOUT','Log');
    local $hbox=new Gtk::HBox 0,1;
    local $messbutt=new Gtk::Button('More');
    if (defined($statline)) {
      local $string=$statline->get();
      $statline->destroy();
      $statline=new Gtk::Label($string);
    } else {
      $statline=new Gtk::Label('');
    }
    $messbutt->signal_connect('clicked',\&show_stdout);
    $hbox->pack_start($messbutt,0,0,0);
    $statline->show;
    $hbox->pack_start($statline,0,0,0);
    $messbutt->show;
    $hbox->show;
    $main_window_vbox->pack_start($hbox,1,1,1);
    $main_window_vbox->show;
    $main_window->add($main_window_vbox);
    $main_window->show;
    $main_window->signal_connect('delete_event', \&main::quit_client);
}

sub show_credits()
{
  local $label=new Gtk::Label('
Yeemp version 0.4.5
Copyright 2001 Nathaniel E. Bray/Deekoo L.
Licensed under the GPL and the Net::SSLeay Documentation License.

This product includes software developed by Sampo Kellomaki (sampo@iki.fi)
');
  local $win=new Gtk::Window;
  $win->add($label);
  $label->show;
  $win->show;
}

sub redraw_main_window()
{
  $main_window->destroy;
  &draw_main_window();
}

sub cb_ynprompt()
{
  &cb_closebutton_3rdgen($_[0]);
  $_[1]($_[2]);
  0;
}

sub shutdown_ui()
{
  # Should Gtk be deinitialized in any way?
}

sub dec_openwindows()
{
	$openwindows--;
  if (!defined($Yeemp::Config::maxopenwindows)) {
    return;
  }
	&UI::debug("Open down to $openwindows/$Yeemp::Config::maxopenwindows\n");
	if ($openwindows==$Yeemp::Config::maxopenwindows-1) {
		&main::show_new_messages();
	}
}

sub ynprompt($$$;$)
{
  # $_[0]: Question.
  # $_[1]: Yes function.
  # $_[2]: Data for yes function.
  # $_[3]: Optional no function.
  local $win=new Gtk::Window;
	$openwindows++;
  local $vbox=new Gtk::VBox (0,1);
  local $label=new Gtk::Label($_[0]);
  local $hbox=new Gtk::HBox(0,1);
  local $yes=new Gtk::Button($locale{"Yes"});
  local $no=new Gtk::Button($locale{"No"});
  $yes->signal_connect('clicked', \&cb_ynprompt, $_[1],$_[2]);
  if (defined($_[3])) {
    $no->signal_connect('clicked', \&cb_ynprompt, $_[3], $_[2]);
  } else {
    $no->signal_connect('clicked', \&cb_closebutton_3rdgen);
  }
  $hbox->pack_start($yes,0,0,5);
  $hbox->pack_end($no,0,0,5);
  $vbox->pack_start($label,0,0,5);
  $vbox->pack_start($hbox,0,0,5);
  $win->add($vbox);
  $yes->show;
  $no->show;
  $hbox->show;
  $label->show;
  $vbox->show;
	$win->signal_connect('destroy', \&dec_openwindows);
  $win->show;
}

sub cb_del_contact()
{
  &ynprompt("Are you sure you want to remove\n$_[1] from your contact list?",
    \&cb_del_contact_confirmed, $_[1]);
  0;
}

sub cb_edit_contact_ok()
{
  local $contact=$_[1];
	$edit_contact_win{$_[1]}->destroy;
	local $shortname=$edit_contact_name{$_[1]}->get_text;
	if ($shortname ne '') {
	  $Yeemp::Client::contacts_shortnames{$contact}=$shortname;
	}
	if (defined($edit_presence{$_[1]})) {
		$Yeemp::Client::contacts_presence{$_[1]} =
			$edit_presence{$_[1]};
	} else {
		delete($Yeemp::Client::contacts_presence{$_[1]});
	}
  0;
}

sub show_authmenu_help()
{
  my $help = new Gtk::Label($locale{authmenu_help});
  my $win = new Gtk::Window;
  $win->add($help);
  $help->show;
  $win->show;
}

sub cb_purge_contact($)
{
  &Yeemp::Client::purge_contact($_[0]);
  $edit_contact_win{$_[0]}->destroy;
  &redraw_main_window();
  0;
}

sub cb_edit_contact()
{
  $edit_contact_win{$_[1]}=new Gtk::Window;
	local $vbox=new Gtk::VBox(0,1);
	local $label=new Gtk::Label("Editing $_[1]");
	$vbox->pack_start($label,0,0,0);
	local $ext;
	if (-f $Yeemp::Client::clientdir."cache/$_[1]/portrait.xpm") {
		$ext = "xpm";
	} elsif (-f $Yeemp::Client::clientdir."cache/$_[1]/portrait.png") {
	  $ext = "png";
	} elsif (-f $Yeemp::Client::clientdir."cache/$_[1]/portrait.jpg") {
		$ext = 'jpg';
	}
	if (defined($ext)) {
		if ($ext ne 'xpm') {
			system("${ext}topnm ${Yeemp::Client::clientdir}cache/$_[1]/portrait.$ext |ppmtoxpm >${Yeemp::Client::clientdir}cache/$_[1]/portrait.xpm");
		}
	  $edit_contact_win{$_[1]}->realize;
		local @pixmap=Gtk::Gdk::Pixmap->create_from_xpm(
			$edit_contact_win{$_[1]}->window, Gtk::Gdk::Color->parse_color('black'),
				"$Yeemp::Client::clientdir/cache/$_[1]/portrait.xpm");
#			`${ext}topnm ${Yeemp::Client::clientdir}cache/$_[1]/portrait.$ext | ppmtoxpm`);
		local $pixmapwid = new Gtk::Pixmap @pixmap;
		$pixmapwid->show;
		$vbox->pack_start($pixmapwid,0,0,0);
		&UI::debug("neep.\n");
	}
	$label->show;
	local $namebox=new Gtk::HBox(0,1);
	local $namelabel=new Gtk::Label("Name:");
	$namebox->pack_start($namelabel,0,0,0);
	$namelabel->show;
	$edit_contact_name{$_[1]}=new Gtk::Entry;
	if (exists($Yeemp::Client::contacts_shortnames{$_[1]})) {
		$edit_contact_name{$_[1]}->set_text($Yeemp::Client::contacts_shortnames{$_[1]});
	}
	$namebox->pack_end($edit_contact_name{$_[1]},0,0,0);
	$edit_contact_name{$_[1]}->show;
	$vbox->pack_start($namebox,0,0,0);
	$namebox->show;
	local $authbox=new Gtk::HBox(0,1);
	local $authlabel = new Gtk::Label('Presence authorization:');
	$authbox->pack_start($authlabel,0,0,0);
	$authlabel->show;
	local $authmenu = new Gtk::Menu;
	local $authdefault = new Gtk::RadioMenuItem('Default',undef);
	$authmenu->append($authdefault);
	$authdefault->signal_connect('activate',
		sub { delete($edit_presence{$_[1]}); }, $_[1]);
	$authdefault->show;
	local $authshow = new Gtk::RadioMenuItem('Show', $authdefault);
	$authmenu->append($authshow);
	$authshow->signal_connect('activate',
		sub { $edit_presence{$_[1]}='show'; }, $_[1]);
	$authshow->show;
	local $authconfirm = new Gtk::RadioMenuItem('Confirm', $authshow);
	$authmenu->append($authconfirm);
	$authconfirm->signal_connect('activate',
		sub { $edit_presence{$_[1]}='confirm'; }, $_[1]);
	$authconfirm->show;
	local $authhide = new Gtk::RadioMenuItem('Hide', $authconfirm);
	$authmenu->append($authhide);
	$authhide->signal_connect('activate',
		sub { $edit_presence{$_[1]}='hide'; }, $_[1]);
  $authhide->show;
	if (defined($Yeemp::Client::contacts_presence{$_[1]})) {
	  if ($Yeemp::Client::contacts_presence{$_[1]} eq 'show') {
		  $authmenu->set_active(1);
		} elsif ($Yeemp::Client::contacts_presence{$_[1]} eq 'confirm') {
		  $authmenu->set_active(2);
		} elsif ($Yeemp::Client::contacts_presence{$_[1]} eq 'hide') {
			$authmenu->set_active(3);
		} else {
			&stderr("Unhandled presence mode $Yeemp::Config::contacts_presence{$_[1]}\n");
		}
	}
	local $authoptmenu = new Gtk::OptionMenu();
	$authoptmenu->set_menu($authmenu);
	$authbox->pack_start($authoptmenu,0,0,0);
	$authmenu->show;
	$authoptmenu->show;
	local $authhelp=new Gtk::Button("Help");
	$authhelp->signal_connect('clicked',\&show_authmenu_help);
	$authbox->pack_end($authhelp,0,0,0);
	$authhelp->show;
	$vbox->pack_start($authbox,0,0,0);
	$authbox->show;
	local $purge = new Gtk::Button('Delete this contact');
	$purge->signal_connect('clicked', sub {
			 &ynprompt("This will erase all keys associated with this client, and all unsent messages\nto them.  Are you absolutely certain you want to purge this contact? ", \&cb_purge_contact, $_[1]);
		}, $_[1]);
	$vbox->pack_start($purge,0,0,0);
	$purge->show;
	local $bbox=new Gtk::HBox(0,1);
	local $ok=new Gtk::Button("OK");
	$bbox->pack_start($ok,0,0,0);
	$ok->signal_connect('clicked',\&cb_edit_contact_ok, $_[1]);
	$ok->show;
	local $cancel=new Gtk::Button("Cancel");
	$cancel->signal_connect('clicked', sub { &cb_closebutton_3rdgen($_[0]); });
	$bbox->pack_start($cancel,0,0,0);
	$cancel->show;
	$bbox->show;
	$vbox->pack_start($bbox,0,0,0);
	$vbox->show;
	$edit_contact_win{$_[1]}->add($vbox);
	$edit_contact_win{$_[1]}->show;
  0;
}

sub cb_private_status()
{
  &UI::stderr("cb_private_status() not yet written.\n");
  0;
}

sub cb_send_message($$$)
{
  # Arg 0: Widget.
  # Arg 1: Contactstruct
  # Arg 2: Text Widgeta
  my $message = "From: $Yeemp::Config::my_address\012".
    (exists(${$_[1]}{Subto}) && (defined(${$_[1]}{Subto})) ? "Subto: ${$_[1]}{Subto}\012" : '').
    "\012". $_[0]."\012";
  Yeemp::Client::deliver_message(${$_[1]}{To},$message);
  0;
}

sub add_contact_to_box()
{
    local $eventbox=new Gtk::EventBox;
    local $hbox=new Gtk::HBox 0,1;
    &UI::set_statfield($_[1]);
    local $contactname=new Gtk::Label($_[1]);
    local $contact=$_[1];
    $contactname->show;
    $eventbox->signal_connect('event',\&UI::click_handler,$contact);
    $hbox->pack_end($contactname,1,0,0);
    $hbox->pack_start($UI::contacts_statfield{$_[1]},0,0,0);
    $eventbox->add($hbox);
    $_[0]->pack_start($eventbox,0,0,0);
    $hbox->show;
    $eventbox->show;
}

sub cb_add_contact()
{
  # Args:
  # $_[0]: OK button.
  # $_[1]: Address widget
  # $_[2]: Shortname widget
  local $address=${$_[1]}->get_chars(0,-1);
  local $shortname=${$_[2]}->get_chars(0,-1);
  if (length($address)) {
    if (&Yeemp::Client::add_contact($address,$shortname)) {
	  &add_contact_to_box($contactbox,$address);
      &cb_closebutton_3rdgen($_[0]);
	  &redraw_main_window();
    } else {
      &UI::error("$address is already in your contact list.");
    }
  }
  0;
}

sub contact_popup()
{
  # Arg 0: Widget
  # Arg 1: Contact.
  local $widget=$_[0];
  local $popup = new Gtk::Menu;
#  local $send_file = new Gtk::MenuItem('Send File');
#  local $req_file = new Gtk::MenuItem('Request File');
#  local $set_status = new Gtk::MenuItem('Set Status');
  local $edit_contact = new Gtk::MenuItem('Edit Contact');
#  local $del_contact = new Gtk::MenuItem('Delete Contact');
#  $popup->append($send_file);
#  $popup->append($req_file);
#  $popup->append($set_status);
  $popup->append($edit_contact);
#  $send_file->signal_connect('activate', \&cb_send_file, $_[1]);
#  $req_file->signal_connect('activate', \&cb_req_file, $_[1]);
#  $set_status->signal_connect('activate', \&cb_private_status, $_[1]);
#  $del_contact->signal_connect('activate',\&cb_del_contact, $_[1]);
	$edit_contact->signal_connect('activate',\&cb_edit_contact, $_[1]);
#  $send_file->show;
#  $req_file->show;
#  $set_status->show;
  $edit_contact->show;
  $popup->popup(undef,undef,${widget},0);
}

sub message_dialog($)
{
  # Arg 0: ref to hash of To, Subto
  my $prompt = "Message to ${$_[0]}{To}\n";
  if (defined(${$_[0]}{Subto})) {
    $prompt .= "(Subrecipient: ${$_[0]}{Subto})\n";
  }
  read_textarea($prompt, $locale{Send}, $locale{Cancel}, undef,
    \&cb_send_message, $_[0]);
#  my $ta = new UI::TextArea($prompt);
}

sub message_dialog_old($)
{
  # Arg 0: ref to hash of To, Subto
  local $win=new Gtk::Window;
  local $vbox=new Gtk::VBox 0,1;
  local $labelstr = "Message to ${$_[0]}{To}";
  if (defined(${$_[0]}{Subto})) {
    $labelstr.="\n(Subrecipient: ${$_[0]}{Subto})";
  }
  local $label=new Gtk::Label($labelstr);
  local $text=new Gtk::Text;
  local $hbox=new Gtk::HBox 0,1;
  local $ok=new Gtk::Button('Send');
  local $actions=new Gtk::Button('Actions');
  local $cancel=new Gtk::Button('Cancel');
  $vbox->pack_start($label,0,0,0);
  $vbox->pack_start($text,0,0,0);
  $hbox->pack_start($ok,0,0,0);
  $hbox->pack_start($actions,0,0,0);
  $hbox->pack_start($cancel,0,0,0);
  $vbox->pack_start($hbox,0,0,0);
  $text->set_editable(1);
  $ok->signal_connect('clicked', \&cb_send_message, $_[0], \$text);
  $actions->signal_connect('clicked', \&contact_popup, ${$_[0]}{To});
  $cancel->signal_connect('clicked', \&UI::cb_closebutton_3rdgen);
  $win->add($vbox);
  $label->show;
  $text->show;
  $ok->show;
  $actions->show;
  $cancel->show;
  $hbox->show;
  $vbox->show;
  $win->show;
}

sub click_handler()
{
  local %event=%{$_[2]};
  if ($event{type} eq 'button_press') {
    if ($event{button} == 1) {
      local %mhash = ( 'To' => $_[1]);
      &message_dialog(\%mhash);
    } else {
      &contact_popup($_[0],$_[1]);
    }
  }
  return 0;
}

sub show_corrupt_message()
{
	$win = new Gtk::Window;
	$openwindows++;
	$vbox = new Gtk::VBox;
	$text = new Gtk::Text;
	$label = new Gtk::Label('Corrupt message:');
	$vbox->pack_start($label,0,0,0);
	$label->show;
	$text->insert_text(&Yeemp::Client::clean($_[0]));
	$vbox->pack_start($text,0,0,0);
	$close=new Gtk::Button('Dismiss');
	$close->signal_connect('clicked', \&cb_closebutton_2ndgen);
	$vbox->pack_start($close,0,0,0);
	$close->show;
	$win->add($vbox);
	$vbox->show;
	$win->signal_connect('destroy',\&dec_openwindows);
	$win->show;
}

sub show_message()
{
  # Arguments:
  # 0: Reference to message
  # 0: userheader hashref
  # 1: Decrypted message body.
  # 2: Server-supplied headers.
  # 3: Raw encrypted message
  # 4: Decryption errors, if any.
  local $win=new Gtk::Window;
	$openwindows++;
  local $vbox=new Gtk::VBox;
  local $key;
  local $ref=${$_[0]}{userheaders};
  local %headers=%{$_[0]{userheaders}};
  local $header;
  if (defined($_[0]{decrypterrors}) && ($_[0]{decrypterrors} ne '')) {
    local $errorlabel=new Gtk::Label("Warnings: \n$_[0]{decrypterrors}");
    $vbox->pack_start($errorlabel,0,0,0);
    $errorlabel->show;
  }
  my $bodyfield = new UI::StaticText($_[0]{body});
  if ($$bodyfield{wrapped}) {
    my $wrappedlabel = new Gtk::Label("(Text autowrapped)");
    $vbox->pack_start($wrappedlabel, 0, 0, 0);
    $wrappedlabel->show;
  }
  foreach $key (keys(%headers)) {
    $header=new Gtk::Label("$key: $headers{$key}");
	  $vbox->pack_start($header,0,0,0);
    $header->show;
  }
  $vbox->pack_start($$bodyfield{hbox},0,0,0);
  $bodyfield->show;
  local $hbox=new Gtk::HBox;
  local $reply=new Gtk::Button('Reply');
  local $raw=new Gtk::Button('Raw');
  local $close=new Gtk::Button('Close');
  $hbox->pack_start($reply,0,0,0);
  $hbox->pack_start($raw,0,0,0);
  $hbox->pack_start($close,0,0,0);
#  $vbox->pack_start($message,0,0,0);
  $vbox->pack_start($hbox,0,0,0);
  $win->add($vbox);
  $close->signal_connect('clicked', \&cb_closebutton_3rdgen );
  $raw->signal_connect('clicked', \&show_raw_message, $_[0]{serverheaders},
    $_[0]{raw}, $_[0]{decrypterrors});
  local %replyhash = ('To' => $headers{From}, 'Subto' => $headers{Subfrom});
  $reply->signal_connect('clicked', \&cb_reply, \%replyhash);
  if (exists($headers{From})) {
    $reply->show;
  }
  $raw->show;
  $close->show;
  $hbox->show;
#  $message->show;
  $vbox->show;
	$win->signal_connect('destroy', \&dec_openwindows, undef);
  $win->show;
#  $bodyfield->signal_connect('expose_event', sub {
#    $bodyfield->window->draw_pixmap($textcolour, ${$_[1]},
#      0, 0, 0, 0, ${$_[1]}->width, ${$_[1]}->height);
#    my $style = ${$_[0]}->style;
#    $style->draw_string($bodyfield->window(), 'selected', 0, 0, ${$_[1]});
#  }, \$body);
}

sub read_two_strings()
{
  # Args:
  # 0: Header
  # 1: Prompt, string 1
  # 2: Prompt, string 2
  # 3: OK button label
  # 4: Callback function
  local $win=new Gtk::Window;
  local $vbox=new Gtk::VBox(0,1);
  local $title=new Gtk::Label($_[0]);
  local $hbox1=new Gtk::HBox(0,1);
  local $prompt1=new Gtk::Label($_[1]);
  local $ent1=new Gtk::Entry;
  local $hbox2=new Gtk::HBox(0,1);
  local $prompt2=new Gtk::Label($_[2]);
  local $ent2=new Gtk::Entry;
  local $bbox=new Gtk::HBox(0,1);
  local $ok=new Gtk::Button($_[3]);
  local $cancel=new Gtk::Button('Cancel');
  $cancel->signal_connect('clicked',\&cb_closebutton_3rdgen);
  $ok->signal_connect('clicked',$_[4], \$ent1, \$ent2);
  $hbox1->pack_start($prompt1,0,0,5);
  $hbox1->pack_end($ent1,0,0,5);
  $hbox2->pack_start($prompt2,0,0,5);
  $hbox2->pack_end($ent2,0,0,5);
  $bbox->pack_start($ok,0,0,5);
  $bbox->pack_end($cancel,0,0,5);
  $vbox->pack_start($title,0,0,5);
  $vbox->pack_start($hbox1,0,0,0);
  $vbox->pack_start($hbox2,0,0,0);
  $vbox->pack_start($bbox,0,0,5);
  $win->add($vbox);
  $ok->show;
  $cancel->show;
  $bbox->show;
  $ent1->show;
  $prompt1->show;
  $ent2->show;
  $prompt2->show;
  $title->show;
  $hbox1->show;
  $hbox2->show;
  $vbox->show;
  $win->show;
}

sub cb_reply($$)
{
  &message_dialog($_[1]);
  0;
}

sub cb_closebutton_3rdgen()
{
  local $button=$_[0];
  bless($button,'Gtk::Button');
  $button->parent->parent->parent->destroy;
  Gtk->gc;
  0;
}

sub cb_closebutton_2ndgen()
{
  local $button=$_[0];
  bless($button,'Gtk::Button');
  $button->parent->parent->destroy;
  Gtk->gc;
  0;
}

sub show_raw_message()
{
  # Args:
  # 0: Widget
  # 1: Server headers
  # 2: Encrypted message
  # 3: Decryption errors
  local $win=new Gtk::Window;
  local $vbox=new Gtk::VBox;
  local $close=new Gtk::Button('Close');
  $close->signal_connect('clicked',\&cb_closebutton_2ndgen);
  local $errors;
  if (defined($_[3]) && ($_[3] ne '')) {
    $errors=new Gtk::Label("Decryption errors:\n$_[3]");
  } else {
    $errors=new Gtk::Label('(No decryption errors)');
  }
  local $text=new Gtk::Text;
  $text->insert_text(&Yeemp::Client::clean($_[2]));
  $vbox->pack_start($errors,0,0,0);
  $vbox->pack_start($text,0,0,0);
  $vbox->pack_start($close,0,0,0);
  $win->add($vbox);
  $errors->show;
  $text->show;
  $close->show;
  $vbox->show;
  $win->show;
}

sub set_statfield()
{
  # Args:
  # 0: User's address
  local $file="$Yeemp::Client::clientdir/cache/$_[0]/$Yeemp::Client::contacts_status{$_[0]}.xpm";
  if (-f $file) {
    $contacts_statfield{$_[0]}=
      Gtk::Gdk::Pixmap->create_from_xpm($_[0],undef,$file);
  } else {
    $contacts_statfield{$_[0]}=new Gtk::Label($Yeemp::Client::contacts_status{$_[0]});
  }
  $contacts_statfield{$_[0]}->show;
}

sub show_statchange()
{
  # Args: 0: User, 1: Status.
  &stdout("$_[0] changed their status to $_[1]\n");
  local $box=\$contacts_statfield{$_[0]}->parent;
  bless($$box,'Gtk::HBox');
  $contacts_statfield{$_[0]}->destroy;
  &set_statfield($_[0]);
  $$box->pack_end($UI::contacts_statfield{$_[0]},0,0,0);
}

sub cb_grant_pubkey_req()
{
  $widget=shift;
  if (&Yeemp::Client::grant_public_key_request(@_)) {
    &cb_closebutton_3rdgen($widget);
  }
}

sub show_public_key_request()
{
  # $_[0]: From
  # $_[1]: Key
  # $_[2]: Fingerprint of requestor
  local $win=new Gtk::Window;
	$openwindows++;
  local $vbox=new Gtk::VBox(0,1);
  local $label=new Gtk::Label("Public key request received from\n$_[0].\n($_[0]'s key fingerprint is $_[2])\n\nGrant it?");
  $label->show;
  local $bbox=new Gtk::HBox(0,1);
  local $grant=new Gtk::Button('Grant Request');
  $grant->signal_connect('clicked',\&cb_grant_pubkey_req,@_);
  $grant->show;
  local $ignore=new Gtk::Button('Ignore');
  $ignore->signal_connect('clicked', \&cb_closebutton_3rdgen);
  $ignore->show;
  $vbox->pack_start($label,0,0,0);
  $bbox->pack_start($grant,0,0,0);
  $bbox->pack_end($ignore,0,0,0);
  $bbox->show;
  $vbox->pack_start($bbox,0,0,0);
  $vbox->show;
  $win->add($vbox);
	$win->signal_connect('destroy', \&dec_openwindows);
  $win->show;
}

sub edit_contact_by_address()
{
  # Arg 0: Address
  &UI::stderr("Can't edit contact in GUI yet.\n");
}

sub cb_use_public_key()
{
  local ($address, $key, $msgid) = @{$_[0]};
  &Yeemp::Client::import_contact_pubkey($address,$key);
}

sub cb_discard_public_key()
{
  local ($address, $key, $msgid) = @{$_[0]};
}

sub show_inbound_public_key()
{
  # 0: Sender
  # 1: Key
  # 2: Message-ID of key
  local @arr=($_[0], $_[1], $_[2]);
  &ynprompt("Public key received from $_[0]\n(fingerprint: $Yeemp::Client::fingerprint[0])\nUse it?",
    \&cb_use_public_key, \@arr, \&cb_discard_public_key); 
}

sub cb_set_prefs()
{
  my %pref = %{$_[1]};
  my $address=$pref{address}->get_text();
  if ($address ne '') {
    &UI::debug("Address set to $address\n");
    $Yeemp::Config::my_address=$address;
  }
  my $server_name=$pref{server}->get_text();
  if ($server_name ne '') {
    &UI::debug("Server name set to $server_name\n");
    $Yeemp::Config::server_name=$server_name;
  }
	$Yeemp::Config::maxportraitsize = $pref{maxportraitsize}->get_value_as_int;
  $Yeemp::Config::use_esd = $pref{use_esd}->get_active() || 0;
  $Yeemp::Config::play_oggs = $pref{play_oggs}->get_active() || 0;
  &cb_closebutton_3rdgen($_[0]);
  undef(%Yeemp::Config::offline);
  &Yeemp::Client::save_client_cfg();
  main_quit Gtk;
}

sub cb_nukebutton_3rdgen()
{
  main_quit Gtk;
  &cb_closebutton_3rdgen($_[0]);
}

sub cb_prefs_copy_login_key()
{
  local $server=$server_name_field->get_text();
  if ($server ne '') {
    $server=$Yeemp::Config::server_name;
  }
  local $address=$my_address_field->get_text();
  if ($address ne '') {
    $address=$Yeemp::Config::my_address;
  }
  &Yeemp::Client::copy_login_key($server,$address);
}

sub init_preferences()
{
  my %prefbox;
  $prefbox{win}=new Gtk::Window;
  $prefbox{win}->activate_focus();
  local $vbox=new Gtk::VBox(0,0);
  $prefbox{win}->add($vbox);
  $prefbox{address}=new Gtk::Entry;
  local $header_label=new Gtk::Label("Yeemp Preferences");
  local $address_label=new Gtk::Label('Your Yeemp address: ');
  local $addr_hbox=new Gtk::HBox(0,0);
  $addr_hbox->pack_start($address_label,0,0,0);
  $addr_hbox->pack_start($prefbox{address},1,1,3);
  $vbox->pack_start($header_label,0,0,0);
  $vbox->pack_start($addr_hbox,0,0,0);
  if (defined($Yeemp::Config::my_address)) {
    $prefbox{address}->set_text($Yeemp::Config::my_address);
  }
  local $server_hbox=new Gtk::HBox(0,0);
  local $server_label=new Gtk::Label('Your Yeemp server: ');
  $prefbox{server}=new Gtk::Entry;
  if (defined($Yeemp::Config::server_name)) {
    $prefbox{server}->set_text($Yeemp::Config::server_name);
  } else {
    $Yeemp::Config::offline{$Yeemp::Config::server_name}=1;
  }
  $server_hbox->pack_start($server_label,0,0,0);
  $server_hbox->pack_start($prefbox{server},1,0,0);
  $vbox->pack_start($server_hbox,0,0,0);
	local $portrait_hbox = new Gtk::HBox(0,0);
	$maxportraitsize_label = new Gtk::Label('Max portrait size:');
	local $adj = new Gtk::Adjustment($Yeemp::Config::maxportraitsize, 128,
		2048, 1, 32, 32);
	$prefbox{maxportraitsize} = new Gtk::SpinButton($adj, 0, 0);
	$prefbox{maxportraitsize}->set_usize(75,0);
	$portrait_hbox->pack_start($maxportraitsize_label,0,0,0);
	$portrait_hbox->pack_end($prefbox{maxportraitsize}, 0, 0, 0);
	$maxportraitsize_label->show;
	$prefbox{maxportraitsize}->show;
	$vbox->pack_start($portrait_hbox,0,0,0);
	$portrait_hbox->show;
  $prefbox{play_oggs} = new Gtk::CheckButton("Play oggs");
  $vbox->pack_start($prefbox{play_oggs}, 0, 0, 0);
  $prefbox{play_oggs}->active($Yeemp::Config::play_oggs);
  $prefbox{play_oggs}->show;
  $prefbox{use_esd} = new Gtk::CheckButton("Use ESD if present?");
  $vbox->pack_start($prefbox{use_esd}, 0, 0, 0);
  if (!defined($Yeemp::Config::use_esd)) {
    $Yeemp::Config::use_esd = 0;
  }
  $prefbox{use_esd}->active($Yeemp::Config::use_esd);
  $prefbox{use_esd}->show;
  local $copyloginkey=new Gtk::Button('Copy login key to server');
  $copyloginkey->signal_connect('clicked', \&cb_prefs_copy_login_key);
  $vbox->pack_start($copyloginkey,0,0,0);
  local $buttonbox=new Gtk::HBox(0,0);
  local $ok=new Gtk::Button('Apply');
  local $cancel=new Gtk::Button('Cancel');
  $cancel->signal_connect('clicked', \&cb_nukebutton_3rdgen);
  $ok->signal_connect('clicked', \&cb_set_prefs, \%prefbox);
  $buttonbox->pack_start($cancel,0,0,0);
  $buttonbox->pack_end($ok,0,0,0);
  $vbox->pack_start($buttonbox,0,0,0);
  $cancel->show;
  $ok->show;
  $buttonbox->show;
  $copyloginkey->show;
  $prefbox{server}->show;
  $server_label->show;
  $server_hbox->show;
  $address_label->show;
  $prefbox{address}->show;
  $addr_hbox->show;
  $header_label->show;
  $vbox->show;
  $prefbox{win}->show;
  main Gtk;
}

1;
